@@ -44,6 +44,15 @@
 
 #include <pipewire/pipewire.h>
 
+#define ATOMIC_INC(s)                   __atomic_add_fetch(&(s), 1, __ATOMIC_SEQ_CST)
+#define ATOMIC_LOAD(s)                  __atomic_load_n(&(s), __ATOMIC_SEQ_CST)
+
+#define SEQ_WRITE(s)                    ATOMIC_INC(s)
+#define SEQ_WRITE_SUCCESS(s1,s2)        ((s1) + 1 == (s2) && ((s2) & 1) == 0)
+
+#define SEQ_READ(s)                     ATOMIC_LOAD(s)
+#define SEQ_READ_SUCCESS(s1,s2)         ((s1) == (s2) && ((s2) & 1) == 0)
+
 PW_LOG_TOPIC_STATIC(alsa_log_topic, "alsa.pcm");
 #define PW_LOG_TOPIC_DEFAULT alsa_log_topic
 
@@ -55,9 +64,32 @@ PW_LOG_TOPIC_STATIC(alsa_log_topic, "alsa.pcm");
 
 #define MIN_PERIOD	64
 
+#define MIN_PERIOD_BYTES	(128)
+#define MAX_PERIOD_BYTES	(2*1024*1024)
+
+#define MIN_BUFFER_BYTES	(2*MIN_PERIOD_BYTES)
+#define MAX_BUFFER_BYTES	(2*MAX_PERIOD_BYTES)
+
+struct params {
+	const char *node_name;
+	const char *server_name;
+	const char *playback_node;
+	const char *capture_node;
+	const char *role;
+	snd_pcm_format_t format;
+	int rate;
+	int channels;
+	int period_bytes;
+	int buffer_bytes;
+	uint32_t flags;
+};
+
 typedef struct {
 	snd_pcm_ioplug_t io;
 
+	snd_output_t *output;
+	FILE *log_file;
+
 	char *node_name;
 	char *target;
 	char *role;
@@ -69,6 +101,7 @@ typedef struct {
 	unsigned int draining:1;
 	unsigned int xrun_detected:1;
 	unsigned int hw_params_changed:1;
+	unsigned int active:1;
 
 	snd_pcm_uframes_t hw_ptr;
 	snd_pcm_uframes_t boundary;
@@ -89,53 +122,58 @@ typedef struct {
 	struct pw_stream *stream;
 	struct spa_hook stream_listener;
 
-	struct pw_time time;
+	int64_t delay;
+	uint64_t transfered;
+	uint64_t buffered;
+	int64_t now;
+	uintptr_t seq;
 
 	struct spa_audio_info_raw format;
 } snd_pcm_pipewire_t;
 
 static int snd_pcm_pipewire_stop(snd_pcm_ioplug_t *io);
 
-static int block_check(snd_pcm_ioplug_t *io)
+static int check_active(snd_pcm_ioplug_t *io)
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
 	snd_pcm_sframes_t avail;
-	uint64_t val;
+	bool active;
 
 	avail = snd_pcm_ioplug_avail(io, pw->hw_ptr, io->appl_ptr);
-	if (avail >= 0 && avail < (snd_pcm_sframes_t)pw->min_avail) {
-		spa_system_eventfd_read(pw->system, io->poll_fd, &val);
-		return 1;
-	}
-	return 0;
-}
-
-static int pcm_poll_block_check(snd_pcm_ioplug_t *io)
-{
-	snd_pcm_pipewire_t *pw = io->private_data;
 
 	if (io->state == SND_PCM_STATE_DRAINING) {
-		uint64_t val;
-		spa_system_eventfd_read(pw->system, io->poll_fd, &val);
-		return 0;
-	} else if (io->state == SND_PCM_STATE_RUNNING ||
-			   (io->state == SND_PCM_STATE_PREPARED && io->stream == SND_PCM_STREAM_CAPTURE)) {
-		return block_check(io);
+		active = pw->drained;
 	}
-	return 0;
+	else if (avail >= 0 && avail < (snd_pcm_sframes_t)pw->min_avail) {
+		active = false;
+	}
+	else if (avail >= (snd_pcm_sframes_t)pw->min_avail) {
+		active = true;
+	} else {
+		active = false;
+	}
+	if (pw->active != active) {
+		pw_log_trace("%p: avail:%lu min-avail:%lu state:%s hw:%lu appl:%lu active:%d->%d state:%s",
+			pw, avail, pw->min_avail, snd_pcm_state_name(io->state),
+			pw->hw_ptr, io->appl_ptr, pw->active, active,
+			snd_pcm_state_name(io->state));
+	}
+	return active;
 }
 
-static inline int pcm_poll_unblock_check(snd_pcm_ioplug_t *io)
+
+static int update_active(snd_pcm_ioplug_t *io)
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
-	snd_pcm_uframes_t avail;
+	pw->active = check_active(io);
+	uint64_t val;
 
-	avail = snd_pcm_ioplug_avail(io, pw->hw_ptr, io->appl_ptr);
-	if (avail >= pw->min_avail || io->state == SND_PCM_STATE_DRAINING) {
-		spa_system_eventfd_write(pw->system, pw->fd, 1);
-		return 1;
-	}
-	return 0;
+	if (pw->active || pw->error < 0)
+		spa_system_eventfd_write(pw->system, io->poll_fd, 1);
+	else
+		spa_system_eventfd_read(pw->system, io->poll_fd, &val);
+
+	return pw->active;
 }
 
 static void snd_pcm_pipewire_free(snd_pcm_pipewire_t *pw)
@@ -143,7 +181,7 @@ static void snd_pcm_pipewire_free(snd_pcm_pipewire_t *pw)
 	if (pw == NULL)
 		return;
 
-	pw_log_debug("%p:", pw);
+	pw_log_debug("%p: free", pw);
 	if (pw->main_loop)
 		pw_thread_loop_stop(pw->main_loop);
 	if (pw->stream)
@@ -156,13 +194,16 @@ static void snd_pcm_pipewire_free(snd_pcm_pipewire_t *pw)
 		pw_thread_loop_destroy(pw->main_loop);
 	free(pw->node_name);
 	free(pw->target);
+	free(pw->role);
+	snd_output_close(pw->output);
+	fclose(pw->log_file);
 	free(pw);
 }
 
 static int snd_pcm_pipewire_close(snd_pcm_ioplug_t *io)
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
-	pw_log_debug("%p:", pw);
+	pw_log_debug("%p: close", pw);
 	snd_pcm_pipewire_free(pw);
 	return 0;
 }
@@ -170,7 +211,7 @@ static int snd_pcm_pipewire_close(snd_pcm_ioplug_t *io)
 static int snd_pcm_pipewire_poll_descriptors(snd_pcm_ioplug_t *io, struct pollfd *pfds, unsigned int space)
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
-	pcm_poll_unblock_check(io); /* unblock socket for polling if needed */
+	update_active(io);
 	pfds->fd = pw->fd;
 	pfds->events = POLLIN | POLLERR | POLLNVAL;
 	return 1;
@@ -188,8 +229,10 @@ static int snd_pcm_pipewire_poll_revents(snd_pcm_ioplug_t *io,
 		return pw->error;
 
 	*revents = pfds[0].revents & ~(POLLIN | POLLOUT);
-	if (pfds[0].revents & POLLIN && !pcm_poll_block_check(io))
+	if (pfds[0].revents & POLLIN && check_active(io)) {
 		*revents |= (io->stream == SND_PCM_STREAM_PLAYBACK) ? POLLOUT : POLLIN;
+		update_active(io);
+	}
 
 	return 0;
 }
@@ -213,29 +256,40 @@ static snd_pcm_sframes_t snd_pcm_pipewire_pointer(snd_pcm_ioplug_t *io)
 static int snd_pcm_pipewire_delay(snd_pcm_ioplug_t *io, snd_pcm_sframes_t *delayp)
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
-	int64_t elapsed = 0, filled, avail;
+	uintptr_t seq1, seq2;
+	int64_t elapsed = 0, delay, now, avail;
+	struct timespec ts;
+	int64_t diff;
 
-	if (pw->time.rate.num != 0) {
-		struct timespec ts;
-		int64_t diff;
+	do {
+		seq1 = SEQ_READ(pw->seq);
+
+		delay = pw->delay + pw->transfered;
+		now = pw->now;
+		if (io->stream == SND_PCM_STREAM_PLAYBACK)
+			avail = snd_pcm_ioplug_hw_avail(io, pw->hw_ptr, io->appl_ptr);
+		else
+			avail = snd_pcm_ioplug_avail(io, pw->hw_ptr, io->appl_ptr);
+
+		seq2 = SEQ_READ(pw->seq);
+	} while (!SEQ_READ_SUCCESS(seq1, seq2));
+
+	if (now != 0 && (io->state == SND_PCM_STATE_RUNNING ||
+	    io->state == SND_PCM_STATE_DRAINING)) {
 		clock_gettime(CLOCK_MONOTONIC, &ts);
-		diff = SPA_TIMESPEC_TO_NSEC(&ts) - pw->time.now;
-	        elapsed = (pw->time.rate.denom * diff) / (pw->time.rate.num * SPA_NSEC_PER_SEC);
+		diff = SPA_TIMESPEC_TO_NSEC(&ts) - now;
+		elapsed = (io->rate * diff) / SPA_NSEC_PER_SEC;
+
+		if (io->stream == SND_PCM_STREAM_PLAYBACK)
+			delay -= SPA_MIN(elapsed, delay);
+		else
+			delay += SPA_MIN(elapsed, (int64_t)io->buffer_size);
 	}
-	if (io->stream == SND_PCM_STREAM_PLAYBACK)
-		avail = snd_pcm_ioplug_hw_avail(io, pw->hw_ptr, io->appl_ptr);
-	else
-		avail = snd_pcm_ioplug_avail(io, pw->hw_ptr, io->appl_ptr);
 
-	filled = pw->time.delay + avail;
+	*delayp = delay + avail;
 
-	if (io->stream == SND_PCM_STREAM_PLAYBACK)
-		*delayp = filled - SPA_MIN(elapsed, filled);
-	else
-		*delayp = filled + elapsed;
-
-	pw_log_trace("avail:%"PRIi64" filled %"PRIi64" elapsed:%"PRIi64" delay:%ld %lu %lu",
-			avail, filled, elapsed, *delayp, pw->hw_ptr, io->appl_ptr);
+	pw_log_trace("avail:%"PRIi64" filled %"PRIi64" elapsed:%"PRIi64" delay:%ld hw:%lu appl:%lu",
+			avail, delay, elapsed, *delayp, pw->hw_ptr, io->appl_ptr);
 
 	return 0;
 }
@@ -388,21 +442,19 @@ static void on_stream_process(void *data)
 	snd_pcm_ioplug_t *io = &pw->io;
 	struct pw_buffer *b;
 	snd_pcm_uframes_t hw_avail, before, want, xfer;
+	struct pw_time pwt;
+	int64_t delay;
 
-	pw_stream_get_time_n(pw->stream, &pw->time, sizeof(pw->time));
+	pw_stream_get_time_n(pw->stream, &pwt, sizeof(pwt));
 
-	if (pw->time.rate.num != 0) {
-		pw->time.delay = pw->time.delay * io->rate * pw->time.rate.num / pw->time.rate.denom;
-		pw->time.rate.denom = io->rate;
-		pw->time.rate.num = 1;
-	}
+	delay = pwt.delay;
+	if (pwt.rate.num != 0)
+		delay = delay * io->rate * pwt.rate.num / pwt.rate.denom;
 
 	before = hw_avail = snd_pcm_ioplug_hw_avail(io, pw->hw_ptr, io->appl_ptr);
 
-	if (pw->drained) {
-		pcm_poll_unblock_check(io); /* unblock socket for polling if needed */
-		return;
-	}
+	if (pw->drained)
+		goto done;
 
 	b = pw_stream_dequeue_buffer(pw->stream);
 	if (b == NULL)
@@ -410,15 +462,28 @@ static void on_stream_process(void *data)
 
 	want = b->requested ? b->requested : hw_avail;
 
+	SEQ_WRITE(pw->seq);
+
+	if (pw->now != pwt.now) {
+		pw->transfered = pw->buffered;
+		pw->buffered = 0;
+	}
+
 	xfer = snd_pcm_pipewire_process(pw, b, &hw_avail, want);
 
-	pw_log_trace("%p: avail-before:%lu avail:%lu want:%lu xfer:%lu",
-			pw, before, hw_avail, want, xfer);
-
+	pw->delay = delay;
+	/* the buffer is now queued in the stream and consumed */
 	if (io->stream == SND_PCM_STREAM_PLAYBACK)
-		pw->time.delay += xfer;
-	else
-		pw->time.delay -= SPA_MIN(pw->time.delay, (int64_t)xfer);
+		pw->transfered += xfer;
+
+	/* more then requested data transfered, use them in next iteration */
+	pw->buffered = (want == 0 || pw->transfered < want) ?  0 : (pw->transfered % want);
+
+	pw->now = pwt.now;
+	SEQ_WRITE(pw->seq);
+
+	pw_log_trace("%p: avail-before:%lu avail:%lu want:%lu xfer:%lu hw:%lu appl:%lu",
+			pw, before, hw_avail, want, xfer, pw->hw_ptr, io->appl_ptr);
 
 	pw_stream_queue_buffer(pw->stream, b);
 
@@ -431,7 +496,8 @@ static void on_stream_process(void *data)
 			pw->drained = false;
 		}
 	}
-	pcm_poll_unblock_check(io); /* unblock socket for polling if needed */
+done:
+	update_active(io);
 }
 
 static const struct pw_stream_events stream_events = {
@@ -484,6 +550,8 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 	if (snd_pcm_sw_params_current(io->pcm, swparams) == 0) {
 		snd_pcm_sw_params_get_avail_min(swparams, &pw->min_avail);
 		snd_pcm_sw_params_get_boundary(swparams, &pw->boundary);
+		snd_pcm_sw_params_dump(swparams, pw->output);
+		fflush(pw->log_file);
 	} else {
 		pw->min_avail = io->period_size;
 		pw->boundary = io->buffer_size;
@@ -492,17 +560,14 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 	min_period = (MIN_PERIOD * io->rate / 48000);
 	pw->min_avail = SPA_MAX(pw->min_avail, min_period);
 
-	pw_log_debug("%p: prepare %d %p %lu %ld", pw,
-			pw->error, pw->stream, io->period_size, pw->min_avail);
+	pw_log_debug("%p: prepare error:%d stream:%p buffer-size:%lu "
+			"period-size:%lu min-avail:%ld", pw, pw->error,
+			pw->stream, io->buffer_size, io->period_size, pw->min_avail);
+
 	if (pw->error >= 0 && pw->stream != NULL && !pw->hw_params_changed)
 		goto done;
 	pw->hw_params_changed = false;
 
-	if (pw->stream != NULL) {
-		pw_stream_destroy(pw->stream);
-		pw->stream = NULL;
-	}
-
 	props = pw_properties_new(NULL, NULL);
 	if (props == NULL)
 		goto error;
@@ -528,13 +593,20 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 		pw_properties_get(props, PW_KEY_MEDIA_ROLE) == NULL)
 		pw_properties_setf(props, PW_KEY_MEDIA_ROLE, "%s", pw->role);
 
+	params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &pw->format);
+
+	if (pw->stream != NULL) {
+		pw_stream_update_properties(pw->stream, &props->dict);
+		pw_stream_update_params(pw->stream, params, 1);
+		goto done;
+	}
+
 	pw->stream = pw_stream_new(pw->core, pw->node_name, props);
 	if (pw->stream == NULL)
 		goto error;
 
 	pw_stream_add_listener(pw->stream, &pw->stream_listener, &stream_events, pw);
 
-	params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &pw->format);
 	pw->error = 0;
 
 	pw_stream_connect(pw->stream,
@@ -550,6 +622,7 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 
 done:
 	pw->hw_ptr = 0;
+	pw->now = 0;
 	pw->xrun_detected = false;
 	pw->drained = false;
 	pw->draining = false;
@@ -568,9 +641,8 @@ static int snd_pcm_pipewire_start(snd_pcm_ioplug_t *io)
 	snd_pcm_pipewire_t *pw = io->private_data;
 
 	pw_thread_loop_lock(pw->main_loop);
-	pw_log_debug("%p:", pw);
+	pw_log_debug("%p: start", pw);
 	pipewire_start(pw);
-	block_check(io); /* unblock socket for polling if needed */
 	pw_thread_loop_unlock(pw->main_loop);
 	return 0;
 }
@@ -579,8 +651,8 @@ static int snd_pcm_pipewire_stop(snd_pcm_ioplug_t *io)
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
 
-	pw_log_debug("%p:", pw);
-	pcm_poll_unblock_check(io);
+	pw_log_debug("%p: stop", pw);
+	update_active(io);
 
 	pw_thread_loop_lock(pw->main_loop);
 	if (pw->activated && pw->stream != NULL) {
@@ -593,7 +665,7 @@ static int snd_pcm_pipewire_stop(snd_pcm_ioplug_t *io)
 
 static int snd_pcm_pipewire_pause(snd_pcm_ioplug_t * io, int enable)
 {
-	pw_log_debug("%p:", io);
+	pw_log_debug("%p: pause", io);
 
 	if (enable)
 		snd_pcm_pipewire_stop(io);
@@ -646,6 +718,9 @@ static int snd_pcm_pipewire_hw_params(snd_pcm_ioplug_t * io,
 	snd_pcm_pipewire_t *pw = io->private_data;
 	bool planar;
 
+	snd_pcm_hw_params_dump(params, pw->output);
+	fflush(pw->log_file);
+
 	pw_log_debug("%p: hw_params buffer_size:%lu period_size:%lu", pw, io->buffer_size, io->period_size);
 
 	switch(io->access) {
@@ -726,6 +801,7 @@ static int snd_pcm_pipewire_sw_params(snd_pcm_ioplug_t * io,
 {
 	snd_pcm_pipewire_t *pw = io->private_data;
 
+	pw_thread_loop_lock(pw->main_loop);
 	if (pw->stream) {
 		snd_pcm_uframes_t min_avail;
 		snd_pcm_sw_params_get_avail_min( sw_params, &min_avail);
@@ -746,6 +822,7 @@ static int snd_pcm_pipewire_sw_params(snd_pcm_ioplug_t * io,
 	} else {
 		pw_log_debug("%p: sw_params pre-prepare noop", pw);
 	}
+	pw_thread_loop_unlock(pw->main_loop);
 
 	return 0;
 }
@@ -797,10 +874,9 @@ static const struct chmap_info chmap_info[] = {
 
 static enum snd_pcm_chmap_position channel_to_chmap(enum spa_audio_channel channel)
 {
-	uint32_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(chmap_info); i++)
-		if (chmap_info[i].channel == channel)
-			return chmap_info[i].pos;
+	SPA_FOR_EACH_ELEMENT_VAR(chmap_info, info)
+		if (info->channel == channel)
+			return info->pos;
 	return SND_CHMAP_UNKNOWN;
 }
 
@@ -820,7 +896,10 @@ static int snd_pcm_pipewire_set_chmap(snd_pcm_ioplug_t * io,
 	pw->format.channels = map->channels;
 	for (i = 0; i < map->channels; i++) {
 		pw->format.position[i] = chmap_to_channel(map->pos[i]);
-		pw_log_debug("map %d: %d %d", i, map->pos[i], pw->format.position[i]);
+		pw_log_debug("map %d: %s / %s", i,
+				snd_pcm_chmap_name(map->pos[i]),
+				spa_debug_type_find_short_name(spa_type_audio_channel,
+					pw->format.position[i]));
 	}
 	return 1;
 }
@@ -890,8 +969,7 @@ static snd_pcm_ioplug_callback_t pipewire_pcm_callback = {
 	.query_chmaps = snd_pcm_pipewire_query_chmaps,
 };
 
-static int pipewire_set_hw_constraint(snd_pcm_pipewire_t *pw, int rate,
-		snd_pcm_format_t format, int channels, int period_bytes)
+static int pipewire_set_hw_constraint(snd_pcm_pipewire_t *pw, struct params *p)
 {
 	unsigned int access_list[] = {
 		SND_PCM_ACCESS_MMAP_INTERLEAVED,
@@ -923,26 +1001,38 @@ static int pipewire_set_hw_constraint(snd_pcm_pipewire_t *pw, int rate,
 	int max_channels;
 	int min_period_bytes;
 	int max_period_bytes;
+	int min_buffer_bytes;
+	int max_buffer_bytes;
 	int err;
 
-	if (rate > 0) {
-		min_rate = max_rate = rate;
+	if (p->rate > 0) {
+		min_rate = max_rate = SPA_CLAMP(p->rate, 1, MAX_RATE);
 	} else {
 		min_rate = 1;
 		max_rate = MAX_RATE;
 	}
-	if (channels > 0) {
-		min_channels = max_channels = channels;
+	if (p->channels > 0) {
+		min_channels = max_channels = SPA_CLAMP(p->channels, 1, MAX_CHANNELS);
 	} else {
 		min_channels = 1;
 		max_channels = MAX_CHANNELS;
 	}
-	if (period_bytes > 0) {
-		min_period_bytes = max_period_bytes = period_bytes;
+	if (p->period_bytes > 0) {
+		min_period_bytes = max_period_bytes = SPA_CLAMP(p->period_bytes,
+				MIN_PERIOD_BYTES, MAX_PERIOD_BYTES);
 	} else {
-		min_period_bytes = 128;
-		max_period_bytes = 2*1024*1024;
+		min_period_bytes = MIN_PERIOD_BYTES;
+		max_period_bytes = MAX_PERIOD_BYTES;
 	}
+	if (p->buffer_bytes > 0) {
+		min_buffer_bytes = max_buffer_bytes = SPA_CLAMP(p->buffer_bytes,
+				MIN_BUFFER_BYTES, MAX_BUFFER_BYTES);
+	} else {
+		min_buffer_bytes = MIN_BUFFER_BYTES;
+		max_buffer_bytes = MAX_BUFFER_BYTES;
+	}
+	if (min_period_bytes * 2 > max_buffer_bytes)
+		min_period_bytes = max_period_bytes = max_buffer_bytes / 2;
 
 	if ((err = snd_pcm_ioplug_set_param_list(&pw->io, SND_PCM_IOPLUG_HW_ACCESS,
 						   SPA_N_ELEMENTS(access_list), access_list)) < 0 ||
@@ -951,22 +1041,22 @@ static int pipewire_set_hw_constraint(snd_pcm_pipewire_t *pw, int rate,
 		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_RATE,
 						   min_rate, max_rate)) < 0 ||
 		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_BUFFER_BYTES,
-						   MIN_BUFFERS*min_period_bytes,
-						   MIN_BUFFERS*max_period_bytes)) < 0 ||
+						   min_buffer_bytes,
+						   max_buffer_bytes)) < 0 ||
 		(err = snd_pcm_ioplug_set_param_minmax(&pw->io,
 						   SND_PCM_IOPLUG_HW_PERIOD_BYTES,
 						   min_period_bytes,
 						   max_period_bytes)) < 0 ||
 		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_PERIODS,
-						   MIN_BUFFERS, MAX_BUFFERS)) < 0) {
+						   MIN_BUFFERS, 1024)) < 0) {
 		pw_log_warn("Can't set param list: %s", snd_strerror(err));
 		return err;
 	}
 
-	if (format != SND_PCM_FORMAT_UNKNOWN) {
+	if (p->format != SND_PCM_FORMAT_UNKNOWN) {
 		err = snd_pcm_ioplug_set_param_list(&pw->io,
 				SND_PCM_IOPLUG_HW_FORMAT,
-				1, (unsigned int *)&format);
+				1, (unsigned int *)&p->format);
 		if (err < 0) {
 			pw_log_warn("Can't set param list: %s", snd_strerror(err));
 			return err;
@@ -994,7 +1084,7 @@ static void on_core_error(void *data, uint32_t id, int seq, int res, const char
 	if (id == PW_ID_CORE) {
 		pw->error = res;
 		if (pw->fd != -1)
-			pcm_poll_unblock_check(&pw->io);
+			update_active(&pw->io);
 	}
 	pw_thread_loop_signal(pw->main_loop, false);
 }
@@ -1004,51 +1094,96 @@ static const struct pw_core_events core_events = {
 	.error = on_core_error,
 };
 
-static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
-				const char *node_name,
-				const char *server_name,
-				const char *playback_node,
-				const char *capture_node,
-				const char *role,
-				snd_pcm_stream_t stream,
-				int mode,
-				uint32_t flags,
-				int rate,
-				snd_pcm_format_t format,
-				int channels,
-				int period_bytes)
+
+static ssize_t log_write(void *cookie, const char *buf, size_t size)
+{
+	int len;
+
+	while (size > 0) {
+		len = strcspn(buf, "\n");
+		if (len > 0)
+			pw_log_debug("%.*s", (int)len, buf);
+		buf += len + 1;
+		size -= len + 1;
+	}
+	return size;
+}
+
+static cookie_io_functions_t io_funcs = {
+	.write = log_write,
+};
+
+static int snd_pcm_pipewire_open(snd_pcm_t **pcmp,
+		struct params *p, snd_pcm_stream_t stream, int mode)
 {
 	snd_pcm_pipewire_t *pw;
 	int err;
 	const char *str;
 	struct pw_properties *props = NULL;
 	struct pw_loop *loop;
+	uint32_t val;
 
 	assert(pcmp);
 	pw = calloc(1, sizeof(*pw));
 	if (!pw)
 		return -ENOMEM;
 
+	props = pw_properties_new(NULL, NULL);
+	if (props == NULL) {
+		err = -errno;
+		goto error;
+	}
+
+	str = getenv("PIPEWIRE_ALSA");
+	if (str != NULL) {
+		pw_properties_update_string(props, str, strlen(str));
+		if ((str = pw_properties_get(props, "alsa.format")))
+			p->format = snd_pcm_format_value(str);
+		if ((str = pw_properties_get(props, "alsa.rate")) &&
+		    spa_atou32(str, &val, 0))
+			p->rate = val;
+		if ((str = pw_properties_get(props, "alsa.channels")) &&
+		    spa_atou32(str, &val, 0))
+			p->channels = val;
+		if ((str = pw_properties_get(props, "alsa.period-bytes")) &&
+		    spa_atou32(str, &val, 0))
+			p->period_bytes = val;
+		if ((str = pw_properties_get(props, "alsa.buffer-bytes")) &&
+		    spa_atou32(str, &val, 0))
+			p->buffer_bytes = val;
+	}
+
 	str = getenv("PIPEWIRE_REMOTE");
 	if (str != NULL && str[0] != '\0')
-		server_name = str;
+		p->server_name = str;
 
 	str = getenv("PIPEWIRE_NODE");
 
-	pw_log_debug("%p: open %s %d %d %08x %d %s %d %d '%s'", pw, name,
-			stream, mode, flags, rate,
-			format != SND_PCM_FORMAT_UNKNOWN ? snd_pcm_format_name(format) : "none",
-			channels, period_bytes, str);
+	pw_log_debug("%p: open name:%s stream:%s mode:%d flags:%08x rate:%d format:%s "
+			"channels:%d period-bytes:%d buffer-bytes:%d target:'%s'", pw, p->node_name,
+			snd_pcm_stream_name(stream), mode, p->flags, p->rate,
+			snd_pcm_format_name(p->format), p->channels, p->period_bytes,
+			p->buffer_bytes, str);
 
 	pw->fd = -1;
 	pw->io.poll_fd = -1;
-	pw->flags = flags;
+	pw->flags = p->flags;
+	pw->log_file = fopencookie(pw, "w", io_funcs);
+	if (pw->log_file == NULL) {
+		pw_log_error("can't create log file: %m");
+		err = -errno;
+		goto error;
+	}
+	if ((err = snd_output_stdio_attach(&pw->output, pw->log_file, 0)) < 0) {
+		pw_log_error("can't attach log file: %s", snd_strerror(err));
+		goto error;
+	}
 
-	if (node_name == NULL)
+	if (p->node_name == NULL)
 		pw->node_name = spa_aprintf("ALSA %s",
 			       stream == SND_PCM_STREAM_PLAYBACK ? "Playback" : "Capture");
 	else
-		pw->node_name = strdup(node_name);
+		pw->node_name = strdup(p->node_name);
 
 	if (pw->node_name == NULL) {
 		err = -errno;
@@ -1060,12 +1195,12 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
 		pw->target = strdup(str);
 	else {
 		if (stream == SND_PCM_STREAM_PLAYBACK)
-			pw->target = playback_node ? strdup(playback_node) : NULL;
+			pw->target = p->playback_node ? strdup(p->playback_node) : NULL;
 		else
-			pw->target = capture_node ? strdup(capture_node) : NULL;
+			pw->target = p->capture_node ? strdup(p->capture_node) : NULL;
 	}
 
-	pw->role = (role && *role) ? strdup(role) : NULL;
+	pw->role = (p->role && *p->role) ? strdup(p->role) : NULL;
 
 	pw->main_loop = pw_thread_loop_new("alsa-pipewire", NULL);
 	if (pw->main_loop == NULL) {
@@ -1083,13 +1218,11 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
 		goto error;
 	}
 
-	props = pw_properties_new(NULL, NULL);
-
 	pw_properties_setf(props, PW_KEY_APP_NAME, "PipeWire ALSA [%s]",
 			pw_get_prgname());
 
-	if (server_name)
-		pw_properties_set(props, PW_KEY_REMOTE_NAME, server_name);
+	if (p->server_name)
+		pw_properties_set(props, PW_KEY_REMOTE_NAME, p->server_name);
 
 	if ((err = pw_thread_loop_start(pw->main_loop)) < 0)
 		goto error;
@@ -1121,20 +1254,21 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
 #endif
 	pw->io.flags |= SND_PCM_IOPLUG_FLAG_MONOTONIC;
 
-	if ((err = snd_pcm_ioplug_create(&pw->io, name, stream, mode)) < 0)
+	if ((err = snd_pcm_ioplug_create(&pw->io, p->node_name, stream, mode)) < 0)
 		goto error;
 
-	pw_log_debug("%p: open %s %d %d", pw, name, pw->io.stream, mode);
-
-	if ((err = pipewire_set_hw_constraint(pw, rate, format, channels,
-					period_bytes)) < 0)
+	if ((err = pipewire_set_hw_constraint(pw, p)) < 0)
 		goto error;
 
+	pw_log_debug("%p: opened name:%s stream:%s mode:%d", pw, p->node_name,
+			snd_pcm_stream_name(pw->io.stream), mode);
+
 	*pcmp = pw->io.pcm;
 
 	return 0;
 
 error:
+	pw_log_debug("%p: failed to open %s :%s", pw, p->node_name, spa_strerror(err));
 	pw_properties_free(props);
 	snd_pcm_pipewire_free(pw);
 	return err;
@@ -1145,16 +1279,7 @@ SPA_EXPORT
 SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 {
 	snd_config_iterator_t i, next;
-	const char *node_name = NULL;
-	const char *server_name = NULL;
-	const char *playback_node = NULL;
-	const char *capture_node = NULL;
-	const char *role = NULL;
-	snd_pcm_format_t format = SND_PCM_FORMAT_UNKNOWN;
-	int rate = 0;
-	int channels = 0;
-	int period_bytes = 0;
-	uint32_t flags = 0;
+	struct params params;
 	int err;
 
 	pw_init(NULL, NULL);
@@ -1162,6 +1287,8 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 		return -ENOTSUP;
 
 	PW_LOG_TOPIC_INIT(alsa_log_topic);
+	spa_zero(params);
+	params.format = SND_PCM_FORMAT_UNKNOWN;
 
 	snd_config_for_each(i, next, conf) {
 		snd_config_t *n = snd_config_iterator_entry(i);
@@ -1171,35 +1298,35 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 		if (spa_streq(id, "comment") || spa_streq(id, "type") || spa_streq(id, "hint"))
 			continue;
 		if (spa_streq(id, "name")) {
-			snd_config_get_string(n, &node_name);
+			snd_config_get_string(n, &params.node_name);
 			continue;
 		}
 		if (spa_streq(id, "server")) {
-			snd_config_get_string(n, &server_name);
+			snd_config_get_string(n, &params.server_name);
 			continue;
 		}
 		if (spa_streq(id, "playback_node")) {
-			snd_config_get_string(n, &playback_node);
+			snd_config_get_string(n, &params.playback_node);
 			continue;
 		}
 		if (spa_streq(id, "capture_node")) {
-			snd_config_get_string(n, &capture_node);
+			snd_config_get_string(n, &params.capture_node);
 			continue;
 		}
 		if (spa_streq(id, "role")) {
-			snd_config_get_string(n, &role);
+			snd_config_get_string(n, &params.role);
 			continue;
 		}
 		if (spa_streq(id, "exclusive")) {
 			if (snd_config_get_bool(n))
-				flags |= PW_STREAM_FLAG_EXCLUSIVE;
+				params.flags |= PW_STREAM_FLAG_EXCLUSIVE;
 			continue;
 		}
 		if (spa_streq(id, "rate")) {
 			long val;
 
 			if (snd_config_get_integer(n, &val) == 0)
-				rate = val;
+				params.rate = val;
 			else
 				SNDERR("%s: invalid type", id);
 			continue;
@@ -1208,8 +1335,8 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 			const char *str;
 
 			if (snd_config_get_string(n, &str) == 0) {
-				format = snd_pcm_format_value(str);
-				if (format == SND_PCM_FORMAT_UNKNOWN)
+				params.format = snd_pcm_format_value(str);
+				if (*str && params.format == SND_PCM_FORMAT_UNKNOWN)
 					SNDERR("%s: invalid value %s", id, str);
 			} else {
 				SNDERR("%s: invalid type", id);
@@ -1220,7 +1347,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 			long val;
 
 			if (snd_config_get_integer(n, &val) == 0)
-				channels = val;
+				params.channels = val;
 			else
 				SNDERR("%s: invalid type", id);
 			continue;
@@ -1229,7 +1356,16 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 			long val;
 
 			if (snd_config_get_integer(n, &val) == 0)
-				period_bytes = val;
+				params.period_bytes = val;
+			else
+				SNDERR("%s: invalid type", id);
+			continue;
+		}
+		if (spa_streq(id, "buffer_bytes")) {
+			long val;
+
+			if (snd_config_get_integer(n, &val) == 0)
+				params.buffer_bytes = val;
 			else
 				SNDERR("%s: invalid type", id);
 			continue;
@@ -1238,9 +1374,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 		return -EINVAL;
 	}
 
-	err = snd_pcm_pipewire_open(pcmp, name, node_name, server_name, playback_node,
-			capture_node, role, stream, mode, flags, rate, format,
-			channels, period_bytes);
+	err = snd_pcm_pipewire_open(pcmp, &params, stream, mode);
 
 	return err;
 }
