@@ -50,7 +50,7 @@ struct spa_latency_info {
 	uint64_t max_ns;
 };
 
-#define SPA_LATENCY_INFO(dir,...) (struct spa_latency_info) { .direction = (dir), ## __VA_ARGS__ }
+#define SPA_LATENCY_INFO(dir,...) ((struct spa_latency_info) { .direction = (dir), ## __VA_ARGS__ })
 
 static inline int
 spa_latency_info_compare(const struct spa_latency_info *a, struct spa_latency_info *b)
@@ -70,8 +70,11 @@ spa_latency_info_combine_start(struct spa_latency_info *info, enum spa_direction
 {
 	*info = SPA_LATENCY_INFO(direction,
 			.min_quantum = FLT_MAX,
+			.max_quantum = 0.0f,
 			.min_rate = UINT32_MAX,
-			.min_ns = UINT64_MAX);
+			.max_rate = 0,
+			.min_ns = UINT64_MAX,
+			.max_ns = 0);
 }
 static inline void
 spa_latency_info_combine_finish(struct spa_latency_info *info)
@@ -143,7 +146,7 @@ struct spa_process_latency_info {
 	uint64_t ns;
 };
 
-#define SPA_PROCESS_LATENCY_INFO_INIT(...)	(struct spa_process_latency_info) { __VA_ARGS__ }
+#define SPA_PROCESS_LATENCY_INFO_INIT(...)	((struct spa_process_latency_info) { __VA_ARGS__ })
 
 static inline int
 spa_process_latency_parse(const struct spa_pod *latency, struct spa_process_latency_info *info)
