@@ -34,6 +34,8 @@ void *_acp_log_data;
 
 struct spa_i18n *acp_i18n;
 
+#define DEFAULT_RATE	48000
+
 #define VOLUME_ACCURACY (PA_VOLUME_NORM/100)  /* don't require volume adjustments to be perfectly correct. don't necessarily extend granularity in software unless the differences get greater than this level */
 
 static const uint32_t channel_table[PA_CHANNEL_POSITION_MAX] = {
@@ -67,7 +69,7 @@ static const uint32_t channel_table[PA_CHANNEL_POSITION_MAX] = {
 	[PA_CHANNEL_POSITION_AUX10] = ACP_CHANNEL_START_Aux + 10,
 	[PA_CHANNEL_POSITION_AUX11] = ACP_CHANNEL_START_Aux + 11,
 	[PA_CHANNEL_POSITION_AUX12] = ACP_CHANNEL_START_Aux + 12,
-	[PA_CHANNEL_POSITION_AUX13] = ACP_CHANNEL_START_Aux + 12,
+	[PA_CHANNEL_POSITION_AUX13] = ACP_CHANNEL_START_Aux + 13,
 	[PA_CHANNEL_POSITION_AUX14] = ACP_CHANNEL_START_Aux + 14,
 	[PA_CHANNEL_POSITION_AUX15] = ACP_CHANNEL_START_Aux + 15,
 	[PA_CHANNEL_POSITION_AUX16] = ACP_CHANNEL_START_Aux + 16,
@@ -201,10 +203,22 @@ static void device_free(void *data)
 	pa_hashmap_free(dev->ports);
 }
 
+static inline void channelmap_to_acp(pa_channel_map *m, uint32_t *map)
+{
+	uint32_t i, j;
+	for (i = 0; i < m->channels; i++) {
+		map[i] = channel_pa2acp(m->map[i]);
+		for (j = 0; j < i; j++) {
+			if (map[i] == map[j])
+				map[i] += 32;
+		}
+
+	}
+}
+
 static void init_device(pa_card *impl, pa_alsa_device *dev, pa_alsa_direction_t direction,
 		pa_alsa_mapping *m, uint32_t index)
 {
-	uint32_t i;
 	char **d;
 
 	dev->card = impl;
@@ -219,8 +233,7 @@ static void init_device(pa_card *impl, pa_alsa_device *dev, pa_alsa_direction_t
 	dev->device.format.channels = m->channel_map.channels;
 	pa_cvolume_reset(&dev->real_volume, dev->device.format.channels);
 	pa_cvolume_reset(&dev->soft_volume, dev->device.format.channels);
-	for (i = 0; i < m->channel_map.channels; i++)
-		dev->device.format.map[i]= channel_pa2acp(m->channel_map.map[i]);
+	channelmap_to_acp(&m->channel_map, dev->device.format.map);
 	dev->direction = direction;
 	dev->proplist = pa_proplist_new();
 	pa_proplist_update(dev->proplist, PA_UPDATE_REPLACE, m->proplist);
@@ -244,10 +257,10 @@ static void init_device(pa_card *impl, pa_alsa_device *dev, pa_alsa_direction_t
 			pa_idxset_string_compare_func);
 	if (m->ucm_context.ucm) {
 		dev->ucm_context = &m->ucm_context;
-		if (impl->ucm.alibpref != NULL) {
+		if (impl->ucm.alib_prefix != NULL) {
 			for (d = m->device_strings; *d; d++) {
-				if (pa_startswith(*d, impl->ucm.alibpref)) {
-					size_t plen = strlen(impl->ucm.alibpref);
+				if (pa_startswith(*d, impl->ucm.alib_prefix)) {
+					size_t plen = strlen(impl->ucm.alib_prefix);
 					size_t len = strlen(*d);
 					memmove(*d, (*d) + plen, len - plen + 1);
 					dev->device.flags |= ACP_DEVICE_UCM_DEVICE;
@@ -301,7 +314,7 @@ static int add_pro_profile(pa_card *impl, uint32_t index)
 	snd_pcm_uframes_t try_period_size, try_buffer_size;
 
 	ss.format = PA_SAMPLE_S32LE;
-	ss.rate = 48000;
+	ss.rate = DEFAULT_RATE;
 	ss.channels = 64;
 
 	ap = pa_xnew0(pa_alsa_profile, 1);
@@ -370,10 +383,10 @@ static int add_pro_profile(pa_card *impl, uint32_t index)
 							&try_period_size, &try_buffer_size,
 							0, NULL, NULL, false))) {
 				pa_alsa_init_proplist_pcm(NULL, m->output_proplist, m->output_pcm);
-				snd_pcm_close(m->output_pcm);
-				m->output_pcm = NULL;
+				pa_proplist_setf(m->output_proplist, "clock.name", "api.alsa.%u", index);
+				pa_alsa_close(&m->output_pcm);
 				m->supported = true;
-				pa_channel_map_init_pro(&m->channel_map, m->sample_spec.channels);
+				pa_channel_map_init_auto(&m->channel_map, m->sample_spec.channels, PA_CHANNEL_MAP_AUX);
 			}
 			pa_idxset_put(ap->output_mappings, m, NULL);
 			free(name);
@@ -400,10 +413,10 @@ static int add_pro_profile(pa_card *impl, uint32_t index)
 							&try_period_size, &try_buffer_size,
 							0, NULL, NULL, false))) {
 				pa_alsa_init_proplist_pcm(NULL, m->input_proplist, m->input_pcm);
-				snd_pcm_close(m->input_pcm);
-				m->input_pcm = NULL;
+				pa_proplist_setf(m->input_proplist, "clock.name", "api.alsa.%u", index);
+				pa_alsa_close(&m->input_pcm);
 				m->supported = true;
-				pa_channel_map_init_pro(&m->channel_map, m->sample_spec.channels);
+				pa_channel_map_init_auto(&m->channel_map, m->sample_spec.channels, PA_CHANNEL_MAP_AUX);
 			}
 			pa_idxset_put(ap->input_mappings, m, NULL);
 			free(name);
@@ -613,7 +626,7 @@ static int report_jack_state(snd_mixer_elem_t *melem, unsigned int mask)
 	pa_card *impl = snd_mixer_elem_get_callback_private(melem);
 	snd_hctl_elem_t *elem = snd_mixer_elem_get_private(melem);
 	snd_ctl_elem_value_t *elem_value;
-	bool plugged_in;
+	bool plugged_in, any_input_port_available;
 	void *state;
 	pa_alsa_jack *jack;
 	struct temp_port_avail *tp, *tports;
@@ -724,6 +737,31 @@ static int report_jack_state(snd_mixer_elem_t *melem, unsigned int mask)
 	if (impl->card.active_profile_index != ACP_INVALID_INDEX)
 		active_available = impl->card.profiles[impl->card.active_profile_index]->available;
 
+	/* First round - detect, if we have any input port available.
+	   If the hardware can report the state for all I/O jacks, only speakers
+	   may be plugged in. */
+	any_input_port_available = false;
+	PA_HASHMAP_FOREACH(profile, impl->profiles, state) {
+		pa_device_port *port;
+		void *state2;
+
+		if (profile->profile.flags & ACP_PROFILE_OFF)
+			continue;
+
+		PA_HASHMAP_FOREACH(port, impl->ports, state2) {
+			if (!pa_hashmap_get(port->profiles, profile->profile.name))
+				continue;
+
+			if (port->port.direction == ACP_DIRECTION_CAPTURE &&
+			    port->port.available != ACP_AVAILABLE_NO) {
+				any_input_port_available = true;
+				goto input_port_found;
+			}
+		}
+	}
+input_port_found:
+
+	/* Second round */
 	PA_HASHMAP_FOREACH(profile, impl->profiles, state) {
 		pa_device_port *port;
 		void *state2;
@@ -757,7 +795,7 @@ static int report_jack_state(snd_mixer_elem_t *melem, unsigned int mask)
 
 		if (has_input_port && !has_output_port && found_available_input_port)
 			available = ACP_AVAILABLE_YES;
-		if (has_output_port && !has_input_port && found_available_output_port)
+		if (has_output_port && (!has_input_port || !any_input_port_available) && found_available_output_port)
 			available = ACP_AVAILABLE_YES;
 		if (has_output_port && has_input_port && found_available_output_port && found_available_input_port)
 			available = ACP_AVAILABLE_YES;
@@ -1045,6 +1083,9 @@ static int read_volume(pa_alsa_device *dev)
 	uint32_t i;
 	int res;
 
+	if (!dev->mixer_handle)
+		return 0;
+
 	if ((res = pa_alsa_path_get_volume(dev->mixer_path, dev->mixer_handle, &dev->mapping->channel_map, &r)) < 0)
 		return res;
 
@@ -1076,6 +1117,9 @@ static void set_volume(pa_alsa_device *dev, const pa_cvolume *v)
 
 	dev->real_volume = *v;
 
+	if (!dev->mixer_handle)
+		return;
+
 	/* Shift up by the base volume */
 	pa_sw_cvolume_divide_scalar(&r, &dev->real_volume, dev->base_volume);
 
@@ -1126,6 +1170,9 @@ static int read_mute(pa_alsa_device *dev)
 	bool mute;
 	int res;
 
+	if (!dev->mixer_handle)
+		return 0;
+
 	if ((res = pa_alsa_path_get_mute(dev->mixer_path, dev->mixer_handle, &mute)) < 0)
 		return res;
 
@@ -1144,6 +1191,10 @@ static int read_mute(pa_alsa_device *dev)
 static void set_mute(pa_alsa_device *dev, bool mute)
 {
 	dev->muted = mute;
+
+	if (!dev->mixer_handle)
+		return;
+
 	pa_alsa_path_set_mute(dev->mixer_path, dev->mixer_handle, mute);
 }
 
@@ -1297,7 +1348,6 @@ static int device_disable(pa_card *impl, pa_alsa_mapping *mapping, pa_alsa_devic
 static int device_enable(pa_card *impl, pa_alsa_mapping *mapping, pa_alsa_device *dev)
 {
 	const char *mod_name;
-	bool ignore_dB = false;
 	uint32_t i, port_index;
 	int res;
 
@@ -1314,7 +1364,7 @@ static int device_enable(pa_card *impl, pa_alsa_mapping *mapping, pa_alsa_device
 
 	dev->device.flags |= ACP_DEVICE_ACTIVE;
 
-	find_mixer(impl, dev, NULL, ignore_dB);
+	find_mixer(impl, dev, NULL, impl->ignore_dB);
 
 	/* Synchronize priority values, as it may have changed when setting the profile */
 	for (i = 0; i < impl->card.n_ports; i++) {
@@ -1335,7 +1385,7 @@ static int device_enable(pa_card *impl, pa_alsa_mapping *mapping, pa_alsa_device
 	if (dev->active_port)
 		dev->active_port->port.flags |= ACP_PORT_ACTIVE;
 
-	if ((res = setup_mixer(impl, dev, ignore_dB)) < 0)
+	if ((res = setup_mixer(impl, dev, impl->ignore_dB)) < 0)
 		return res;
 
 	if (dev->read_volume)
@@ -1482,7 +1532,6 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 	struct acp_card *card;
 	const char *s, *profile_set = NULL, *profile = NULL;
 	char device_id[16];
-	bool ignore_dB = false;
 	uint32_t profile_index;
 	int res;
 
@@ -1503,6 +1552,7 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 	impl->use_ucm = true;
 	impl->auto_profile = true;
 	impl->auto_port = true;
+	impl->ignore_dB = false;
 
 	if (props) {
 		if ((s = acp_dict_lookup(props, "api.alsa.use-ucm")) != NULL)
@@ -1510,7 +1560,7 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 		if ((s = acp_dict_lookup(props, "api.alsa.soft-mixer")) != NULL)
 			impl->soft_mixer = spa_atob(s);
 		if ((s = acp_dict_lookup(props, "api.alsa.ignore-dB")) != NULL)
-			ignore_dB = spa_atob(s);
+			impl->ignore_dB = spa_atob(s);
 		if ((s = acp_dict_lookup(props, "device.profile-set")) != NULL)
 			profile_set = s;
 		if ((s = acp_dict_lookup(props, "device.profile")) != NULL)
@@ -1522,7 +1572,7 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 	}
 
 	impl->ucm.default_sample_spec.format = PA_SAMPLE_S16NE;
-	impl->ucm.default_sample_spec.rate = 44100;
+	impl->ucm.default_sample_spec.rate = DEFAULT_RATE;
 	impl->ucm.default_sample_spec.channels = 2;
 	pa_channel_map_init_extend(&impl->ucm.default_channel_map,
 			impl->ucm.default_sample_spec.channels, PA_CHANNEL_MAP_ALSA);
@@ -1558,7 +1608,7 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 		goto error;
 	}
 
-	impl->profile_set->ignore_dB = ignore_dB;
+	impl->profile_set->ignore_dB = impl->ignore_dB;
 
 	pa_alsa_profile_set_probe(impl->profile_set, impl->ucm.mixers,
 			device_id,
@@ -1722,7 +1772,8 @@ static void sync_mixer(pa_alsa_device *d, pa_device_port *port)
 		setting = data->setting;
 	}
 
-	pa_alsa_path_select(d->mixer_path, setting, d->mixer_handle, d->muted);
+	if (d->mixer_handle)
+		pa_alsa_path_select(d->mixer_path, setting, d->mixer_handle, d->muted);
 
 	if (d->set_mute)
 		d->set_mute(d, d->muted);
