@@ -35,6 +35,7 @@
 #include <spa/utils/list.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/json.h>
+#include <spa/utils/result.h>
 #include <spa/utils/string.h>
 #include <spa/node/node.h>
 #include <spa/node/utils.h>
@@ -42,6 +43,7 @@
 #include <spa/node/keys.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/debug/types.h>
+#include <spa/debug/mem.h>
 #include <spa/param/audio/type-info.h>
 #include <spa/param/param.h>
 #include <spa/pod/filter.h>
@@ -57,6 +59,7 @@ struct props {
 	uint32_t n_pos;
 	uint32_t pos[SPA_AUDIO_MAX_CHANNELS];
 	char clock_name[64];
+	unsigned int debug:1;
 };
 
 static void reset_props(struct props *props)
@@ -65,12 +68,12 @@ static void reset_props(struct props *props)
 	props->rate = 0;
 	props->n_pos = 0;
 	strncpy(props->clock_name, DEFAULT_CLOCK_NAME, sizeof(props->clock_name));
+	props->debug = false;
 }
 
 #define DEFAULT_CHANNELS	2
-#define DEFAULT_RATE		44100
+#define DEFAULT_RATE		48000
 
-#define MAX_SAMPLES	8192
 #define MAX_BUFFERS	16
 #define MAX_PORTS	1
 
@@ -107,6 +110,8 @@ struct impl {
 	struct spa_loop *data_loop;
 	struct spa_system *data_system;
 
+	uint32_t quantum_limit;
+
 	struct props props;
 
 	uint64_t info_all;
@@ -278,14 +283,16 @@ static void on_timeout(struct spa_source *source)
 	struct impl *this = source->data;
 	uint64_t expirations, nsec, duration = 10;
 	uint32_t rate;
+	int res;
 
 	spa_log_trace(this->log, "timeout");
 
-	if (spa_system_timerfd_read(this->data_system,
-				this->timer_source.fd, &expirations) < 0) {
-		if (errno == EAGAIN)
-			return;
-		perror("read timerfd");
+	if ((res = spa_system_timerfd_read(this->data_system,
+				this->timer_source.fd, &expirations)) < 0) {
+		if (res != EAGAIN)
+			spa_log_error(this->log, NAME " %p: timerfd error: %s",
+					this, spa_strerror(res));
+		return;
 	}
 
 	nsec = this->next_time;
@@ -538,11 +545,10 @@ impl_node_port_enum_params(void *object, int seq,
 			SPA_PARAM_BUFFERS_buffers, SPA_POD_CHOICE_RANGE_Int(1, 1, MAX_BUFFERS),
 			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(port->blocks),
 			SPA_PARAM_BUFFERS_size,    SPA_POD_CHOICE_RANGE_Int(
-							MAX_SAMPLES * port->bpf,
+							this->quantum_limit * port->bpf,
 							16 * port->bpf,
 							INT32_MAX),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->bpf),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(port->bpf));
 		break;
 	case SPA_PARAM_IO:
 		switch (result.index) {
@@ -607,6 +613,11 @@ port_set_format(struct impl *this,
 		if (spa_format_audio_raw_parse(format, &info.info.raw) < 0)
 			return -EINVAL;
 
+		if (info.info.raw.rate == 0 ||
+		    info.info.raw.channels == 0 ||
+		    info.info.raw.channels > SPA_AUDIO_MAX_CHANNELS)
+			return -EINVAL;
+
 		if (info.info.raw.format == SPA_AUDIO_FORMAT_F32) {
 			port->bpf = 4 * info.info.raw.channels;
 			port->blocks = 1;
@@ -734,9 +745,8 @@ static int impl_node_process(void *object)
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
 	port = &this->port;
-
-	io = port->io;
-	spa_return_val_if_fail(io != NULL, -EIO);
+	if ((io = port->io) == NULL)
+		return -EIO;
 
 	if (io->status != SPA_STATUS_HAVE_DATA)
 		return io->status;
@@ -744,6 +754,20 @@ static int impl_node_process(void *object)
 		io->status = -EINVAL;
 		return io->status;
 	}
+	if (this->props.debug) {
+		struct buffer *b;
+		uint32_t i;
+
+		b = &port->buffers[io->buffer_id];
+		for (i = 0; i < b->outbuf->n_datas; i++) {
+			uint32_t offs, size;
+			struct spa_data *d = b->outbuf->datas;
+
+			offs = SPA_MIN(d->chunk->offset, d->maxsize);
+			size = SPA_MIN(d->maxsize - offs, d->chunk->size);
+			spa_debug_mem(i, SPA_PTROFF(d[i].data, offs, void), SPA_MIN(16u, size));;
+		}
+	}
 	io->status = SPA_STATUS_OK;
 	return SPA_STATUS_HAVE_DATA;
 }
@@ -779,6 +803,13 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 	return 0;
 }
 
+static int do_remove_timer(struct spa_loop *loop, bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct impl *this = user_data;
+	spa_loop_remove_source(this->data_loop, &this->timer_source);
+	return 0;
+}
+
 static int impl_clear(struct spa_handle *handle)
 {
 	struct impl *this;
@@ -787,7 +818,7 @@ static int impl_clear(struct spa_handle *handle)
 
 	this = (struct impl *) handle;
 
-	spa_loop_remove_source(this->data_loop, &this->timer_source);
+	spa_loop_invoke(this->data_loop, do_remove_timer, 0, NULL, 0, true, this);
 	spa_system_close(this->data_system, this->timer_source.fd);
 
 	return 0;
@@ -857,7 +888,6 @@ impl_init(const struct spa_handle_factory *factory,
 		spa_log_error(this->log, "a data_system is needed");
 		return -EINVAL;
 	}
-
 	spa_hook_list_init(&this->hooks);
 
 	this->node.iface = SPA_INTERFACE_INIT(
@@ -905,7 +935,9 @@ impl_init(const struct spa_handle_factory *factory,
 	for (i = 0; info && i < info->n_items; i++) {
 		const char *k = info->items[i].key;
 		const char *s = info->items[i].value;
-		if (spa_streq(k, SPA_KEY_AUDIO_CHANNELS)) {
+		if (spa_streq(k, "clock.quantum-limit")) {
+			spa_atou32(s, &this->quantum_limit, 0);
+		} else if (spa_streq(k, SPA_KEY_AUDIO_CHANNELS)) {
 			this->props.channels = atoi(s);
 		} else if (spa_streq(k, SPA_KEY_AUDIO_RATE)) {
 			this->props.rate = atoi(s);
