@@ -33,6 +33,7 @@
 #include <spa/node/node.h>
 #include <spa/node/utils.h>
 #include <spa/pod/filter.h>
+#include <spa/pod/dynamic.h>
 #include <spa/pod/parser.h>
 #include <spa/debug/types.h>
 
@@ -42,32 +43,19 @@
 #include "modules/spa/spa-node.h"
 #include "client-node.h"
 
-#define NAME "client-node"
+PW_LOG_TOPIC_EXTERN(mod_topic);
+#define PW_LOG_TOPIC_DEFAULT mod_topic
 
 /** \cond */
 
-#define MAX_INPUTS	1024
-#define MAX_OUTPUTS	1024
-
 #define MAX_BUFFERS	64
 #define MAX_METAS	16u
 #define MAX_DATAS	64u
 #define MAX_AREAS	2048
-#define MAX_MIX		128
 
-#define CHECK_IN_PORT_ID(this,d,p)       ((d) == SPA_DIRECTION_INPUT && (p) < MAX_INPUTS)
-#define CHECK_OUT_PORT_ID(this,d,p)      ((d) == SPA_DIRECTION_OUTPUT && (p) < MAX_OUTPUTS)
-#define CHECK_PORT_ID(this,d,p)          (CHECK_IN_PORT_ID(this,d,p) || CHECK_OUT_PORT_ID(this,d,p))
-#define CHECK_FREE_IN_PORT(this,d,p)     (CHECK_IN_PORT_ID(this,d,p) && (this)->in_ports[p] == NULL)
-#define CHECK_FREE_OUT_PORT(this,d,p)    (CHECK_OUT_PORT_ID(this,d,p) && (this)->out_ports[p] == NULL)
-#define CHECK_FREE_PORT(this,d,p)        (CHECK_FREE_IN_PORT (this,d,p) || CHECK_FREE_OUT_PORT (this,d,p))
-#define CHECK_IN_PORT(this,d,p)          (CHECK_IN_PORT_ID(this,d,p) && (this)->in_ports[p])
-#define CHECK_OUT_PORT(this,d,p)         (CHECK_OUT_PORT_ID(this,d,p) && (this)->out_ports[p])
-#define CHECK_PORT(this,d,p)             (CHECK_IN_PORT (this,d,p) || CHECK_OUT_PORT (this,d,p))
-
-#define GET_IN_PORT(this,p)	(this->in_ports[p])
-#define GET_OUT_PORT(this,p)	(this->out_ports[p])
-#define GET_PORT(this,d,p)	(d == SPA_DIRECTION_INPUT ? GET_IN_PORT(this,p) : GET_OUT_PORT(this,p))
+#define CHECK_FREE_PORT(this,d,p)	(p <= pw_map_get_size(&this->ports[d]) && !CHECK_PORT(this,d,p))
+#define CHECK_PORT(this,d,p)		(pw_map_lookup(&this->ports[d], p) != NULL)
+#define GET_PORT(this,d,p)		(pw_map_lookup(&this->ports[d], p))
 
 #define CHECK_PORT_BUFFER(this,b,p)      (b < p->n_buffers)
 
@@ -88,6 +76,11 @@ struct mix {
 	struct buffer buffers[MAX_BUFFERS];
 };
 
+struct params {
+	uint32_t n_params;
+	struct spa_pod **params;
+};
+
 struct port {
 	struct pw_impl_port *port;
 	struct node *node;
@@ -101,8 +94,7 @@ struct port {
 	struct spa_port_info info;
 	struct pw_properties *properties;
 
-	uint32_t n_params;
-	struct spa_pod **params;
+	struct params params;
 
 	unsigned int removed:1;
 	unsigned int destroyed:1;
@@ -128,15 +120,11 @@ struct node {
 	struct spa_source data_source;
 	int writefd;
 
-	uint32_t n_inputs;
-	uint32_t n_outputs;
-	struct port *in_ports[MAX_INPUTS];
-	struct port *out_ports[MAX_OUTPUTS];
+	struct pw_map ports[2];
 
 	struct port dummy;
 
-	uint32_t n_params;
-	struct spa_pod **params;
+	struct params params;
 };
 
 struct impl {
@@ -192,6 +180,32 @@ struct impl {
 #define pw_client_node_resource_port_set_mix_info(r,...)	\
 	pw_client_node_resource(r,port_set_mix_info,1,__VA_ARGS__)
 
+static int update_params(struct params *p, uint32_t n_params, const struct spa_pod **params)
+{
+	uint32_t i;
+	for (i = 0; i < p->n_params; i++)
+		free(p->params[i]);
+	p->n_params = n_params;
+	if (p->n_params == 0) {
+		free(p->params);
+		p->params = NULL;
+	} else {
+		struct spa_pod **np;
+		np = pw_reallocarray(p->params, p->n_params, sizeof(struct spa_pod *));
+		if (np == NULL) {
+			pw_log_error("%p: can't realloc: %m", p);
+			free(p->params);
+			p->params = NULL;
+			p->n_params = 0;
+			return -errno;
+		}
+		p->params = np;
+	}
+	for (i = 0; i < p->n_params; i++)
+		p->params[i] = params[i] ? spa_pod_copy(params[i]) : NULL;
+	return 0;
+}
+
 static int
 do_port_use_buffers(struct impl *impl,
 		    enum spa_direction direction,
@@ -212,12 +226,13 @@ static struct mix *find_mix(struct port *p, uint32_t mix_id)
 		mix_id = 0;
 	else
 		mix_id++;
-	if (mix_id >= MAX_MIX)
-		return NULL;
+
 	len = pw_array_get_len(&p->mix, struct mix);
 	if (mix_id >= len) {
 		size_t need = sizeof(struct mix) * (mix_id + 1 - len);
 		void *ptr = pw_array_add(&p->mix, need);
+		if (ptr == NULL)
+			return NULL;
 		memset(ptr, 0, need);
 	}
 	mix = pw_array_get_unchecked(&p->mix, mix_id, struct mix);
@@ -257,14 +272,14 @@ static void clear_data(struct node *this, struct spa_data *d)
 		id = SPA_PTR_TO_UINT32(d->data);
 		m = pw_mempool_find_id(this->client->pool, id);
 		if (m) {
-			pw_log_debug(NAME " %p: mem %d", impl, m->id);
+			pw_log_debug("%p: mem %d", impl, m->id);
 			pw_memblock_unref(m);
 		}
 		break;
 	}
 	case SPA_DATA_MemFd:
 	case SPA_DATA_DmaBuf:
-		pw_log_debug(NAME " %p: close fd:%d", impl, (int)d->fd);
+		pw_log_debug("%p: close fd:%d", impl, (int)d->fd);
 		close(d->fd);
 		break;
 	default:
@@ -279,7 +294,7 @@ static int clear_buffers(struct node *this, struct mix *mix)
 	for (i = 0; i < mix->n_buffers; i++) {
 		struct buffer *b = &mix->buffers[i];
 
-		spa_log_debug(this->log, NAME" %p: clear buffer %d", this, i);
+		spa_log_debug(this->log, "%p: clear buffer %d", this, i);
 
 		for (j = 0; j < b->buffer.n_datas; j++) {
 			struct spa_data *d = &b->datas[j];
@@ -308,7 +323,7 @@ static int impl_node_enum_params(void *object, int seq,
 {
 	struct node *this = object;
 	uint8_t buffer[1024];
-	struct spa_pod_builder b = { 0 };
+	struct spa_pod_dynamic_builder b;
 	struct spa_result_node_params result;
 	uint32_t count = 0;
 	bool found = false;
@@ -323,10 +338,10 @@ static int impl_node_enum_params(void *object, int seq,
 		struct spa_pod *param;
 
 		result.index = result.next++;
-		if (result.index >= this->n_params)
+		if (result.index >= this->params.n_params)
 			break;
 
-		param = this->params[result.index];
+		param = this->params.params[result.index];
 
 		if (param == NULL || !spa_pod_is_object_id(param, id))
 			continue;
@@ -336,14 +351,15 @@ static int impl_node_enum_params(void *object, int seq,
 		if (result.index < start)
 			continue;
 
-		spa_pod_builder_init(&b, buffer, sizeof(buffer));
-		if (spa_pod_filter(&b, &result.param, param, filter) != 0)
-			continue;
+		spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+		if (spa_pod_filter(&b.b, &result.param, param, filter) == 0) {
+			pw_log_debug("%p: %d param %u", this, seq, result.index);
+			spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+			count++;
+		}
+		spa_pod_dynamic_builder_clean(&b);
 
-		pw_log_debug(NAME " %p: %d param %u", this, seq, result.index);
-		spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
-
-		if (++count == num)
+		if (count == num)
 			break;
 	}
 	return found ? 0 : -ENOENT;
@@ -407,7 +423,7 @@ static int impl_node_send_command(void *object, const struct spa_command *comman
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 	spa_return_val_if_fail(command != NULL, -EINVAL);
 
-	pw_log_debug(NAME " %p: send command %d", this, SPA_COMMAND_TYPE(command));
+	pw_log_debug("%p: send command %d", this, SPA_COMMAND_TYPE(command));
 
 	if (this->resource == NULL)
 		return -EIO;
@@ -429,19 +445,19 @@ static int impl_node_add_listener(void *object,
 {
 	struct node *this = object;
 	struct spa_hook_list save;
-	uint32_t i;
+	union pw_map_item *item;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
 	spa_hook_list_isolate(&this->hooks, &save, listener, events, data);
 
-	for (i = 0; i < MAX_INPUTS; i++) {
-		if (this->in_ports[i])
-			emit_port_info(this, this->in_ports[i]);
+	pw_array_for_each(item, &this->ports[SPA_DIRECTION_INPUT].items) {
+		if (item->data)
+			emit_port_info(this, item->data);
 	}
-	for (i = 0; i < MAX_OUTPUTS; i++) {
-		if (this->out_ports[i])
-			emit_port_info(this, this->out_ports[i]);
+	pw_array_for_each(item, &this->ports[SPA_DIRECTION_OUTPUT].items) {
+		if (item->data)
+			emit_port_info(this, item->data);
 	}
 	spa_hook_list_join(&this->hooks, &save);
 
@@ -469,7 +485,7 @@ impl_node_sync(void *object, int seq)
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	pw_log_debug(NAME " %p: sync", this);
+	pw_log_debug("%p: sync", this);
 
 	if (this->resource == NULL)
 		return -EIO;
@@ -485,17 +501,9 @@ do_update_port(struct node *this,
 	       const struct spa_pod **params,
 	       const struct spa_port_info *info)
 {
-	uint32_t i;
-
 	if (change_mask & PW_CLIENT_NODE_PORT_UPDATE_PARAMS) {
-		spa_log_debug(this->log, NAME" %p: port %u update %d params", this, port->id, n_params);
-		for (i = 0; i < port->n_params; i++)
-			free(port->params[i]);
-		port->n_params = n_params;
-		port->params = realloc(port->params, port->n_params * sizeof(struct spa_pod *));
-		for (i = 0; i < port->n_params; i++) {
-			port->params[i] = params[i] ? spa_pod_copy(params[i]) : NULL;
-		}
+		spa_log_debug(this->log, "%p: port %u update %d params", this, port->id, n_params);
+		update_params(&port->params, n_params, params);
 	}
 
 	if (change_mask & PW_CLIENT_NODE_PORT_UPDATE_INFO) {
@@ -523,7 +531,7 @@ clear_port(struct node *this, struct port *port)
 {
 	struct mix *mix;
 
-	spa_log_debug(this->log, NAME" %p: clear port %p", this, port);
+	spa_log_debug(this->log, "%p: clear port %p", this, port);
 
 	do_update_port(this, port,
 		       PW_CLIENT_NODE_PORT_UPDATE_PARAMS |
@@ -534,18 +542,8 @@ clear_port(struct node *this, struct port *port)
 	pw_array_clear(&port->mix);
 	pw_array_init(&port->mix, sizeof(struct mix) * 2);
 
-	if (port->direction == SPA_DIRECTION_INPUT) {
-		if (this->in_ports[port->id] == port) {
-			this->in_ports[port->id] = NULL;
-			this->n_inputs--;
-		}
-	}
-	else {
-		if (this->out_ports[port->id] == port) {
-			this->out_ports[port->id] = NULL;
-			this->n_outputs--;
-		}
-	}
+	pw_map_insert_at(&this->ports[port->direction], port->id, NULL);
+
 	if (!port->removed)
 		spa_node_emit_port_info(&this->hooks, port->direction, port->id, NULL);
 }
@@ -588,19 +586,19 @@ impl_node_port_enum_params(void *object, int seq,
 	struct node *this = object;
 	struct port *port;
 	uint8_t buffer[1024];
-	struct spa_pod_builder b = { 0 };
+	struct spa_pod_dynamic_builder b;
 	struct spa_result_node_params result;
 	uint32_t count = 0;
 	bool found = false;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 	spa_return_val_if_fail(num != 0, -EINVAL);
-	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
 
 	port = GET_PORT(this, direction, port_id);
+	spa_return_val_if_fail(port != NULL, -EINVAL);
 
-	pw_log_debug(NAME " %p: seq:%d port %d.%d id:%u start:%u num:%u n_params:%d",
-			this, seq, direction, port_id, id, start, num, port->n_params);
+	pw_log_debug("%p: seq:%d port %d.%d id:%u start:%u num:%u n_params:%d",
+			this, seq, direction, port_id, id, start, num, port->params.n_params);
 
 	result.id = id;
 	result.next = 0;
@@ -609,10 +607,10 @@ impl_node_port_enum_params(void *object, int seq,
 		struct spa_pod *param;
 
 		result.index = result.next++;
-		if (result.index >= port->n_params)
+		if (result.index >= port->params.n_params)
 			break;
 
-		param = port->params[result.index];
+		param = port->params.params[result.index];
 
 		if (param == NULL || !spa_pod_is_object_id(param, id))
 			continue;
@@ -622,14 +620,15 @@ impl_node_port_enum_params(void *object, int seq,
 		if (result.index < start)
 			continue;
 
-		spa_pod_builder_init(&b, buffer, sizeof(buffer));
-		if (spa_pod_filter(&b, &result.param, param, filter) < 0)
-			continue;
+		spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+		if (spa_pod_filter(&b.b, &result.param, param, filter) == 0) {
+			pw_log_debug("%p: %d param %u", this, seq, result.index);
+			spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+			count++;
+		}
+		spa_pod_dynamic_builder_clean(&b);
 
-		pw_log_debug(NAME " %p: %d param %u", this, seq, result.index);
-		spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
-
-		if (++count == num)
+		if (count == num)
 			break;
 	}
 	return found ? 0 : -ENOENT;
@@ -646,14 +645,14 @@ impl_node_port_set_param(void *object,
 	struct mix *mix;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
-	if(!CHECK_PORT(this, direction, port_id))
-		return param == NULL ? 0 : -EINVAL;
-
-	pw_log_debug(NAME" %p: port %d.%d set param %s %d", this,
-			direction, port_id,
-			spa_debug_type_find_name(spa_type_param, id), id);
 
 	port = GET_PORT(this, direction, port_id);
+	if(port == NULL)
+		return param == NULL ? 0 : -EINVAL;
+
+	pw_log_debug("%p: port %d.%d set param %s %d", this,
+			direction, port_id,
+			spa_debug_type_find_name(spa_type_param, id), id);
 
 	if (id == SPA_PARAM_Format) {
 		pw_array_for_each(mix, &port->mix)
@@ -680,14 +679,13 @@ static int do_port_set_io(struct impl *impl,
 	uint32_t tag[5] = { impl->node_id, direction, port_id, mix_id, id };
 	struct pw_memmap *mm, *old;
 
-	pw_log_debug(NAME " %p: %s port %d.%d set io %p %zd", this,
+	pw_log_debug("%p: %s port %d.%d set io %p %zd", this,
 			direction == SPA_DIRECTION_INPUT ? "input" : "output",
 			port_id, mix_id, data, size);
 
-	if (!CHECK_PORT(this, direction, port_id))
-		return data == NULL ? 0 : -EINVAL;
-
 	port = GET_PORT(this, direction, port_id);
+	if (port == NULL)
+		return data == NULL ? 0 : -EINVAL;
 
 	if ((mix = find_mix(port, mix_id)) == NULL || !mix->valid)
 		return -EINVAL;
@@ -749,21 +747,27 @@ do_port_use_buffers(struct impl *impl,
 	uint32_t i, j;
 	struct pw_client_node_buffer *mb;
 
-	if (!CHECK_PORT(this, direction, port_id))
+	p = GET_PORT(this, direction, port_id);
+	if (p == NULL)
 		return n_buffers == 0 ? 0 : -EINVAL;
 
-	p = GET_PORT(this, direction, port_id);
+	if (n_buffers > MAX_BUFFERS)
+		return -ENOSPC;
 
-	spa_log_debug(this->log, NAME " %p: %s port %d.%d use buffers %p %u flags:%08x", this,
+
+	spa_log_debug(this->log, "%p: %s port %d.%d use buffers %p %u flags:%08x", this,
 			direction == SPA_DIRECTION_INPUT ? "input" : "output",
 			port_id, mix_id, buffers, n_buffers, flags);
 
-	if (direction == SPA_DIRECTION_OUTPUT)
-		mix_id = SPA_ID_INVALID;
-
 	if ((mix = find_mix(p, mix_id)) == NULL || !mix->valid)
 		return -EINVAL;
 
+	if (direction == SPA_DIRECTION_OUTPUT) {
+		mix_id = SPA_ID_INVALID;
+		if ((mix = find_mix(p, mix_id)) == NULL || !mix->valid)
+			return -EINVAL;
+	}
+
 	clear_buffers(this, mix);
 
 	if (n_buffers > 0) {
@@ -804,9 +808,15 @@ do_port_use_buffers(struct impl *impl,
 		}
 		for (j = 0; j < buffers[i]->n_datas; j++) {
 			struct spa_data *d = buffers[i]->datas;
-			if (d->type == SPA_DATA_MemPtr)
-				endptr = SPA_PTROFF(d->data, d->maxsize, void);
+			if (d->type == SPA_DATA_MemPtr) {
+				if ((m = pw_mempool_find_ptr(impl->context->pool, d->data)) == NULL ||
+				    m != mem)
+					return -EINVAL;
+				endptr = SPA_MAX(endptr, SPA_PTROFF(d->data, d->maxsize, void));
+			}
 		}
+		if (endptr > SPA_PTROFF(baseptr, mem->size, void))
+			return -EINVAL;
 
 		m = pw_mempool_import_block(this->client->pool, mem);
 		if (m == NULL)
@@ -818,7 +828,7 @@ do_port_use_buffers(struct impl *impl,
 		mb[i].mem_id = m->id;
 		mb[i].offset = SPA_PTRDIFF(baseptr, mem->map->ptr);
 		mb[i].size = SPA_PTRDIFF(endptr, baseptr);
-		spa_log_debug(this->log, NAME" %p: buffer %d %d %d %d", this, i, mb[i].mem_id,
+		spa_log_debug(this->log, "%p: buffer %d %d %d %d", this, i, mb[i].mem_id,
 				mb[i].offset, mb[i].size);
 
 		b->buffer.n_metas = SPA_MIN(buffers[i]->n_metas, MAX_METAS);
@@ -869,10 +879,6 @@ do_port_use_buffers(struct impl *impl,
 	}
 	mix->n_buffers = n_buffers;
 
-	if (this->resource->version >= 4)
-		pw_client_node_resource_port_set_mix_info(this->resource,
-						 direction, port_id, mix_id,
-						 mix->peer_id, NULL);
 	return pw_client_node_resource_port_use_buffers(this->resource,
 						 direction, port_id, mix_id, flags,
 						 n_buffers, mb);
@@ -903,7 +909,7 @@ impl_node_port_reuse_buffer(void *object, uint32_t port_id, uint32_t buffer_id)
 	struct node *this = object;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
-	spa_return_val_if_fail(CHECK_OUT_PORT(this, SPA_DIRECTION_OUTPUT, port_id), -EINVAL);
+	spa_return_val_if_fail(CHECK_PORT(this, SPA_DIRECTION_OUTPUT, port_id), -EINVAL);
 
 	spa_log_trace_fp(this->log, "reuse buffer %d", buffer_id);
 
@@ -926,7 +932,7 @@ static int impl_node_process(void *object)
 	n->rt.activation->signal_time = SPA_TIMESPEC_TO_NSEC(&ts);
 
 	if (SPA_UNLIKELY(spa_system_eventfd_write(this->data_system, this->writefd, 1) < 0))
-		spa_log_warn(this->log, NAME" %p: error %m", this);
+		spa_log_warn(this->log, "%p: error %m", this);
 
 	return SPA_STATUS_OK;
 }
@@ -940,7 +946,7 @@ client_node_get_node(void *data,
 	struct node *this = &impl->node;
 	uint32_t new_id = user_data_size;
 
-	pw_log_debug(NAME" %p: bind %u/%u", this, new_id, version);
+	pw_log_debug("%p: bind %u/%u", this, new_id, version);
 
 	impl->bind_node_version = version;
 	impl->bind_node_id = new_id;
@@ -960,21 +966,13 @@ client_node_update(void *data,
 	struct node *this = &impl->node;
 
 	if (change_mask & PW_CLIENT_NODE_UPDATE_PARAMS) {
-		uint32_t i;
-		pw_log_debug(NAME" %p: update %d params", this, n_params);
-
-		for (i = 0; i < this->n_params; i++)
-			free(this->params[i]);
-		this->n_params = n_params;
-		this->params = realloc(this->params, this->n_params * sizeof(struct spa_pod *));
-
-		for (i = 0; i < this->n_params; i++)
-			this->params[i] = params[i] ? spa_pod_copy(params[i]) : NULL;
+		pw_log_debug("%p: update %d params", this, n_params);
+		update_params(&this->params, n_params, params);
 	}
 	if (change_mask & PW_CLIENT_NODE_UPDATE_INFO) {
 		spa_node_emit_info(&this->hooks, info);
 	}
-	pw_log_debug(NAME" %p: got node update", this);
+	pw_log_debug("%p: got node update", this);
 	return 0;
 }
 
@@ -992,10 +990,8 @@ client_node_port_update(void *data,
 	struct port *port;
 	bool remove;
 
-	spa_log_debug(this->log, NAME" %p: got port update change:%08x params:%d",
+	spa_log_debug(this->log, "%p: got port update change:%08x params:%d",
 			this, change_mask, n_params);
-	if (!CHECK_PORT_ID(this, direction, port_id))
-		return -EINVAL;
 
 	remove = (change_mask == 0);
 
@@ -1010,6 +1006,9 @@ client_node_port_update(void *data,
 		struct port *target;
 
 		if (port == NULL) {
+			if (!CHECK_FREE_PORT(this, direction, port_id))
+				return -EINVAL;
+
 			target = &this->dummy;
 			spa_zero(this->dummy);
 			target->direction = direction;
@@ -1030,7 +1029,7 @@ static int client_node_set_active(void *data, bool active)
 {
 	struct impl *impl = data;
 	struct node *this = &impl->node;
-	spa_log_debug(this->log, NAME" %p: active:%d", this, active);
+	spa_log_debug(this->log, "%p: active:%d", this, active);
 	return pw_impl_node_set_active(impl->this.node, active);
 }
 
@@ -1055,13 +1054,12 @@ static int client_node_port_buffers(void *data,
 	struct mix *mix;
 	uint32_t i, j;
 
-	spa_log_debug(this->log, NAME " %p: %s port %d.%d buffers %p %u", this,
+	spa_log_debug(this->log, "%p: %s port %d.%d buffers %p %u", this,
 			direction == SPA_DIRECTION_INPUT ? "input" : "output",
 			port_id, mix_id, buffers, n_buffers);
 
-	spa_return_val_if_fail(CHECK_PORT(this, direction, port_id), -EINVAL);
-
 	p = GET_PORT(this, direction, port_id);
+	spa_return_val_if_fail(p != NULL, -EINVAL);
 
 	if (direction == SPA_DIRECTION_OUTPUT)
 		mix_id = SPA_ID_INVALID;
@@ -1120,7 +1118,7 @@ static void node_on_data_fd_events(struct spa_source *source)
 	struct node *this = source->data;
 
 	if (source->rmask & (SPA_IO_ERR | SPA_IO_HUP)) {
-		spa_log_warn(this->log, NAME" %p: got error", this);
+		spa_log_warn(this->log, "%p: got error", this);
 		return;
 	}
 
@@ -1130,12 +1128,12 @@ static void node_on_data_fd_events(struct spa_source *source)
 
 		if (SPA_UNLIKELY(spa_system_eventfd_read(this->data_system,
 					this->data_source.fd, &cmd) < 0))
-			pw_log_warn(NAME" %p: read failed %m", this);
+			pw_log_warn("%p: read failed %m", this);
 		else if (SPA_UNLIKELY(cmd > 1))
-			pw_log_warn("(%s-%u) client missed %"PRIu64" wakeups",
+			pw_log_info("(%s-%u) client missed %"PRIu64" wakeups",
 				node->name, node->info.id, cmd - 1);
 
-		spa_log_trace_fp(this->log, NAME" %p: got ready", this);
+		spa_log_trace_fp(this->log, "%p: got ready", this);
 		spa_node_call_ready(&this->callbacks, SPA_STATUS_HAVE_DATA);
 	}
 }
@@ -1195,12 +1193,7 @@ node_init(struct node *this,
 
 static int node_clear(struct node *this)
 {
-	uint32_t i;
-
-	for (i = 0; i < this->n_params; i++)
-		free(this->params[i]);
-	free(this->params);
-
+	update_params(&this->params, 0, NULL);
 	return 0;
 }
 
@@ -1222,7 +1215,7 @@ static void client_node_resource_destroy(void *data)
 	struct pw_impl_client_node *this = &impl->this;
 	struct node *node = &impl->node;
 
-	pw_log_debug(NAME " %p: destroy", node);
+	pw_log_debug("%p: destroy", node);
 
 	impl->node.resource = this->resource = NULL;
 	spa_hook_remove(&impl->resource_listener);
@@ -1247,7 +1240,7 @@ static void client_node_resource_error(void *data, int seq, int res, const char
 	struct node *this = &impl->node;
 	struct spa_result_node_error result;
 
-	pw_log_error(NAME " %p: error seq:%d %d (%s)", this, seq, res, message);
+	pw_log_error("%p: error seq:%d %d (%s)", this, seq, res, message);
 	result.message = message;
 	spa_node_emit_result(&this->hooks, seq, res, SPA_RESULT_TYPE_NODE_ERROR, &result);
 }
@@ -1257,7 +1250,7 @@ static void client_node_resource_pong(void *data, int seq)
 	struct impl *impl = data;
 	struct node *this = &impl->node;
 
-	pw_log_debug(NAME " %p: got pong, emit result %d", this, seq);
+	pw_log_debug("%p: got pong, emit result %d", this, seq);
 	spa_node_emit_result(&this->hooks, seq, 0, 0, NULL);
 }
 
@@ -1268,11 +1261,11 @@ void pw_impl_client_node_registered(struct pw_impl_client_node *this, struct pw_
 	struct pw_impl_client *client = impl->node.client;
 	uint32_t node_id = global->id;
 
-	pw_log_debug(NAME " %p: %d", &impl->node, node_id);
+	pw_log_debug("%p: %d", &impl->node, node_id);
 
 	impl->activation = pw_mempool_import_block(client->pool, node->activation);
 	if (impl->activation == NULL) {
-		pw_log_debug(NAME " %p: can't import block: %m", &impl->node);
+		pw_log_debug("%p: can't import block: %m", &impl->node);
 		return;
 	}
 	impl->node_id = node_id;
@@ -1312,7 +1305,7 @@ static void node_initialized(void *data)
 	node->writefd = impl->fds[1];
 
 	spa_loop_add_source(node->data_loop, &node->data_source);
-	pw_log_debug(NAME " %p: transport read-fd:%d write-fd:%d", node, impl->fds[0], impl->fds[1]);
+	pw_log_debug("%p: transport read-fd:%d write-fd:%d", node, impl->fds[0], impl->fds[1]);
 
 	size = sizeof(struct spa_io_buffers) * MAX_AREAS;
 
@@ -1324,7 +1317,7 @@ static void node_initialized(void *data)
 	if (impl->io_areas == NULL)
                 return;
 
-	pw_log_debug(NAME " %p: io areas %p", node, impl->io_areas->map->ptr);
+	pw_log_debug("%p: io areas %p", node, impl->io_areas->map->ptr);
 
 	if ((global = pw_impl_node_get_global(this->node)) != NULL)
 		pw_impl_client_node_registered(this, global);
@@ -1341,7 +1334,7 @@ static void node_free(void *data)
 
 	this->node = NULL;
 
-	pw_log_debug(NAME " %p: free", node);
+	pw_log_debug("%p: free", node);
 	node_clear(node);
 
 	spa_hook_remove(&impl->node_listener);
@@ -1357,6 +1350,8 @@ static void node_free(void *data)
 	if (impl->io_areas)
 		pw_memblock_unref(impl->io_areas);
 
+	pw_map_clear(&impl->node.ports[0]);
+	pw_map_clear(&impl->node.ports[1]);
 	pw_map_clear(&impl->io_map);
 
 	if (impl->fds[0] != -1)
@@ -1376,8 +1371,15 @@ static int port_init_mix(void *data, struct pw_impl_port_mix *mix)
 		return -ENOMEM;
 
 	mix->id = pw_map_insert_new(&impl->io_map, NULL);
-	if (mix->id == SPA_ID_INVALID)
+	if (mix->id == SPA_ID_INVALID) {
+		m->valid = false;
 		return -errno;
+	}
+	if (mix->id > MAX_AREAS) {
+		pw_map_remove(&impl->io_map, mix->id);
+		m->valid = false;
+		return -ENOMEM;
+	}
 
 	mix->io = SPA_PTROFF(impl->io_areas->map->ptr,
 			mix->id * sizeof(struct spa_io_buffers), void);
@@ -1385,7 +1387,7 @@ static int port_init_mix(void *data, struct pw_impl_port_mix *mix)
 
 	m->peer_id = mix->peer_id;
 
-	pw_log_debug(NAME " %p: init mix id:%d io:%p base:%p", impl,
+	pw_log_debug("%p: init mix id:%d io:%p base:%p", impl,
 			mix->id, mix->io, impl->io_areas->map->ptr);
 
 	return 0;
@@ -1398,7 +1400,7 @@ static int port_release_mix(void *data, struct pw_impl_port_mix *mix)
 	struct node *this = &impl->node;
 	struct mix *m;
 
-	pw_log_debug(NAME " %p: remove mix id:%d io:%p base:%p",
+	pw_log_debug("%p: remove mix id:%d io:%p base:%p",
 			this, mix->id, mix->io, impl->io_areas->map->ptr);
 
 	if ((m = find_mix(port, mix->port.port_id)) == NULL || !m->valid)
@@ -1445,7 +1447,7 @@ impl_mix_add_port(void *object, enum spa_direction direction, uint32_t mix_id,
 		const struct spa_dict *props)
 {
 	struct port *port = object;
-	pw_log_debug(NAME " %p: add port %d:%d.%d", object, direction, port->id, mix_id);
+	pw_log_debug("%p: add port %d:%d.%d", object, direction, port->id, mix_id);
 	return 0;
 }
 
@@ -1453,7 +1455,7 @@ static int
 impl_mix_remove_port(void *object, enum spa_direction direction, uint32_t mix_id)
 {
 	struct port *port = object;
-	pw_log_debug(NAME " %p: remove port %d:%d.%d", object, direction, port->id, mix_id);
+	pw_log_debug("%p: remove port %d:%d.%d", object, direction, port->id, mix_id);
 	return 0;
 }
 
@@ -1478,6 +1480,7 @@ static int impl_mix_port_set_io(void *object,
 	struct port *p = object;
 	struct pw_impl_port *port = p->port;
 	struct impl *impl = port->owner_data;
+	struct node *this = &impl->node;
 	struct pw_impl_port_mix *mix;
 
 	mix = pw_map_lookup(&port->mix_port_map, mix_id);
@@ -1489,6 +1492,11 @@ static int impl_mix_port_set_io(void *object,
 			mix->io = data;
 		else
 			mix->io = NULL;
+
+		if (mix->io != NULL && this->resource && this->resource->version >= 4)
+			pw_client_node_resource_port_set_mix_info(this->resource,
+						 direction, port->port_id,
+						 mix->port.port_id, mix->peer_id, NULL);
 	}
 
 	return do_port_set_io(impl,
@@ -1526,7 +1534,7 @@ static void node_port_init(void *data, struct pw_impl_port *port)
 	struct port *p = pw_impl_port_get_user_data(port);
 	struct node *this = &impl->node;
 
-	pw_log_debug(NAME " %p: port %p init", this, port);
+	pw_log_debug("%p: port %p init", this, port);
 
 	*p = this->dummy;
 	p->port = port;
@@ -1541,13 +1549,7 @@ static void node_port_init(void *data, struct pw_impl_port *port)
 			&impl_port_mix, p);
 	ensure_mix(impl, p, SPA_ID_INVALID);
 
-	if (p->direction == SPA_DIRECTION_INPUT) {
-		this->in_ports[p->id] = p;
-		this->n_inputs++;
-	} else {
-		this->out_ports[p->id] = p;
-		this->n_outputs++;
-	}
+	pw_map_insert_at(&this->ports[p->direction], p->id, p);
 	return;
 }
 
@@ -1556,14 +1558,14 @@ static void node_port_added(void *data, struct pw_impl_port *port)
 	struct impl *impl = data;
 	struct port *p = pw_impl_port_get_user_data(port);
 
-	pw_impl_port_set_mix(port, &p->mix_node,
-			PW_IMPL_PORT_MIX_FLAG_MULTI |
-			PW_IMPL_PORT_MIX_FLAG_MIX_ONLY);
-
 	port->flags |= PW_IMPL_PORT_FLAG_NO_MIXER;
 
 	port->impl = SPA_CALLBACKS_INIT(&port_impl, p);
 	port->owner_data = impl;
+
+	pw_impl_port_set_mix(port, &p->mix_node,
+			PW_IMPL_PORT_MIX_FLAG_MULTI |
+			PW_IMPL_PORT_MIX_FLAG_MIX_ONLY);
 }
 
 static void node_port_removed(void *data, struct pw_impl_port *port)
@@ -1572,7 +1574,7 @@ static void node_port_removed(void *data, struct pw_impl_port *port)
 	struct node *this = &impl->node;
 	struct port *p = pw_impl_port_get_user_data(port);
 
-	pw_log_debug(NAME " %p: port %p remove", this, port);
+	pw_log_debug("%p: port %p remove", this, port);
 
 	p->removed = true;
 	clear_port(this, p);
@@ -1589,10 +1591,10 @@ static void node_peer_added(void *data, struct pw_impl_node *peer)
 
 	m = pw_mempool_import_block(this->client->pool, peer->activation);
 	if (m == NULL) {
-		pw_log_debug(NAME " %p: can't ensure mem: %m", this);
+		pw_log_debug("%p: can't ensure mem: %m", this);
 		return;
 	}
-	pw_log_debug(NAME " %p: peer %p id:%u added mem_id:%u", &impl->this, peer,
+	pw_log_debug("%p: peer %p id:%u added mem_id:%u", &impl->this, peer,
 			peer->info.id, m->id);
 
 	if (this->resource == NULL)
@@ -1617,11 +1619,11 @@ static void node_peer_removed(void *data, struct pw_impl_node *peer)
 
 	m = pw_mempool_find_fd(this->client->pool, peer->activation->fd);
 	if (m == NULL) {
-		pw_log_warn(NAME " %p: unknown peer %p fd:%d", this, peer,
+		pw_log_warn("%p: unknown peer %p fd:%d", this, peer,
 			peer->source.fd);
 		return;
 	}
-	pw_log_debug(NAME " %p: peer %p %u removed", this, peer,
+	pw_log_debug("%p: peer %p %u removed", this, peer,
 			peer->info.id);
 
 	if (this->resource != NULL) {
@@ -1641,7 +1643,7 @@ static void node_driver_changed(void *data, struct pw_impl_node *old, struct pw_
 	struct impl *impl = data;
 	struct node *this = &impl->node;
 
-	pw_log_debug(NAME " %p: driver changed %p -> %p", this, old, driver);
+	pw_log_debug("%p: driver changed %p -> %p", this, old, driver);
 
 	node_peer_removed(data, old);
 	node_peer_added(data, driver);
@@ -1708,7 +1710,7 @@ struct pw_impl_client_node *pw_impl_client_node_new(struct pw_resource *resource
 
 	impl->context = context;
 	impl->fds[0] = impl->fds[1] = -1;
-	pw_log_debug(NAME " %p: new", &impl->node);
+	pw_log_debug("%p: new", &impl->node);
 
 	support = pw_context_get_support(impl->context, &n_support);
 	node_init(&impl->node, NULL, support, n_support);
@@ -1717,6 +1719,8 @@ struct pw_impl_client_node *pw_impl_client_node_new(struct pw_resource *resource
 	impl->node.client = client;
 	this->flags = do_register ? 0 : 1;
 
+	pw_map_init(&impl->node.ports[0], 64, 64);
+	pw_map_init(&impl->node.ports[1], 64, 64);
 	pw_map_init(&impl->io_map, 64, 64);
 
 	this->resource = resource;
