@@ -46,7 +46,8 @@
 #include "client-node.h"
 #include "transport.h"
 
-#define NAME "client-node0"
+PW_LOG_TOPIC_EXTERN(mod_topic);
+#define PW_LOG_TOPIC_DEFAULT mod_topic
 
 /** \cond */
 
@@ -301,7 +302,7 @@ static int impl_node_enum_params(void *object, int seq,
 		if (spa_pod_filter(&b, &result.param, param, filter) != 0)
 			continue;
 
-		pw_log_debug(NAME " %p: %d param %u", this, seq, result.index);
+		pw_log_debug("%p: %d param %u", this, seq, result.index);
 		spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
 
 		if (++count == num)
@@ -359,7 +360,7 @@ static int send_clock_update(struct node *this)
 
 	clock_gettime(CLOCK_MONOTONIC, &ts);
 	now = SPA_TIMESPEC_TO_NSEC(&ts);
-	pw_log_trace(NAME " %p: now %"PRIi64, this, now);
+	pw_log_trace("%p: now %"PRIi64, this, now);
 
 	struct spa_command_node0_clock_update cu =
 		SPA_COMMAND_NODE0_CLOCK_UPDATE_INIT(type,
@@ -419,7 +420,7 @@ impl_node_sync(void *object, int seq)
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
-	pw_log_debug(NAME " %p: sync %p", this, this->resource);
+	pw_log_debug("%p: sync %p", this, this->resource);
 
 	if (this->resource == NULL)
 		return -EIO;
@@ -470,8 +471,19 @@ do_update_port(struct node *this,
 		for (i = 0; i < port->n_params; i++)
 			free(port->params[i]);
 		port->n_params = n_params;
-		port->params = realloc(port->params, port->n_params * sizeof(struct spa_pod *));
-
+		if (port->n_params == 0) {
+			free(port->params);
+			port->params = NULL;
+		} else {
+			void *p;
+			p = pw_reallocarray(port->params, port->n_params, sizeof(struct spa_pod *));
+			if (p == NULL) {
+				pw_log_error("%p: port %u can't realloc: %m", this, port_id);
+				free(port->params);
+				port->n_params = 0;
+			}
+			port->params = p;
+		}
 		for (i = 0; i < port->n_params; i++) {
 			port->params[i] = params[i] ?
 				pw_protocol_native0_pod_from_v2(this->resource->client, params[i]) : NULL;
@@ -482,8 +494,7 @@ do_update_port(struct node *this,
 	}
 
 	if (change_mask & PW_CLIENT_NODE0_PORT_UPDATE_INFO) {
-		if (port->properties)
-			pw_properties_free(port->properties);
+		pw_properties_free(port->properties);
 		port->properties = NULL;
 		port->info.props = NULL;
 		port->info.n_params = 0;
@@ -579,7 +590,7 @@ impl_node_port_enum_params(void *object, int seq,
 
 	port = GET_PORT(this, direction, port_id);
 
-	pw_log_debug(NAME " %p: %d port %d.%d %u %u %u", this, seq,
+	pw_log_debug("%p: %d port %d.%d %u %u %u", this, seq,
 			direction, port_id, id, start, num);
 
 	result.id = id;
@@ -606,7 +617,7 @@ impl_node_port_enum_params(void *object, int seq,
 		if (spa_pod_filter(&b, &result.param, param, filter) < 0)
 			continue;
 
-		pw_log_debug(NAME " %p: %d param %u", this, seq, result.index);
+		pw_log_debug("%p: %d param %u", this, seq, result.index);
 		spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
 
 		if (++count == num)
@@ -770,7 +781,7 @@ impl_node_port_use_buffers(void *object,
 
 		mb[i].buffer = &b->buffer;
 		mb[i].mem_id = b->memid;
-		mb[i].offset = SPA_PTRDIFF(baseptr, SPA_MEMBER(mem->map->ptr, mem->map->offset, void));
+		mb[i].offset = SPA_PTRDIFF(baseptr, SPA_PTROFF(mem->map->ptr, mem->map->offset, void));
 		mb[i].size = data_size;
 
 		for (j = 0; j < buffers[i]->n_metas; j++)
@@ -1033,8 +1044,19 @@ client_node0_update(void *data,
 		for (i = 0; i < this->n_params; i++)
 			free(this->params[i]);
 		this->n_params = n_params;
-		this->params = realloc(this->params, this->n_params * sizeof(struct spa_pod *));
-
+		if (this->n_params == 0) {
+			free(this->params);
+			this->params = NULL;
+		} else {
+			void *p;
+			p = pw_reallocarray(this->params, this->n_params, sizeof(struct spa_pod *));
+			if (p == NULL) {
+				pw_log_error("%p: can't realloc: %m", this);
+				free(this->params);
+				this->n_params = 0;
+			}
+			this->params = p;
+		}
 		for (i = 0; i < this->n_params; i++)
 			this->params[i] = params[i] ? spa_pod_copy(params[i]) : NULL;
 	}
@@ -1096,7 +1118,7 @@ static void client_node0_event(void *data, struct spa_event *event)
 	}
 }
 
-static struct pw_client_node0_methods client_node0_methods = {
+static const struct pw_client_node0_methods client_node0_methods = {
 	PW_VERSION_CLIENT_NODE0_METHODS,
 	.done = client_node0_done,
 	.update = client_node0_update,
@@ -1304,19 +1326,19 @@ static const struct pw_resource_events resource_events = {
 
 static void convert_properties(struct pw_properties *properties)
 {
-	struct {
+	static const struct {
 		const char *from, *to;
 	} props[] = {
 		{ "pipewire.autoconnect", PW_KEY_NODE_AUTOCONNECT, },
 		{ "pipewire.target.node", PW_KEY_NODE_TARGET, }
 	};
-	uint32_t i;
+
 	const char *str;
 
-	for(i = 0; i < SPA_N_ELEMENTS(props); i++) {
-		if ((str = pw_properties_get(properties, props[i].from)) != NULL) {
-			pw_properties_set(properties, props[i].to, str);
-			pw_properties_set(properties, props[i].from, NULL);
+	SPA_FOR_EACH_ELEMENT_VAR(props, p) {
+		if ((str = pw_properties_get(properties, p->from)) != NULL) {
+			pw_properties_set(properties, p->to, str);
+			pw_properties_set(properties, p->from, NULL);
 		}
 	}
 }
@@ -1342,7 +1364,6 @@ struct pw_impl_client_node0 *pw_impl_client_node0_new(struct pw_resource *resour
 	const struct spa_support *support;
 	uint32_t n_support;
 	const char *name;
-	const char *str;
 	int res;
 
 	impl = calloc(1, sizeof(struct impl));
@@ -1388,8 +1409,7 @@ struct pw_impl_client_node0 *pw_impl_client_node0_new(struct pw_resource *resour
 		goto error_no_node;
 	}
 
-	str = pw_properties_get(properties, "pipewire.client.reuse");
-	impl->client_reuse = str && pw_properties_parse_bool(str);
+	impl->client_reuse = pw_properties_get_bool(properties, "pipewire.client.reuse", false);
 
 	pw_resource_add_listener(this->resource,
 				 &impl->resource_listener,
