@@ -159,32 +159,28 @@ uint32_t format_name2id(const char *name)
 
 uint32_t format_paname2id(const char *name, size_t size)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_formats); i++) {
-		if (audio_formats[i].name != NULL &&
-		    strncmp(name, audio_formats[i].name, size) == 0)
-			return audio_formats[i].id;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_formats, f) {
+		if (f->name != NULL &&
+		    strncmp(name, f->name, size) == 0)
+			return f->id;
 	}
 	return SPA_AUDIO_FORMAT_UNKNOWN;
 }
 
 enum sample_format format_id2pa(uint32_t id)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_formats); i++) {
-		if (id == audio_formats[i].id)
-			return audio_formats[i].pa;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_formats, f) {
+		if (id == f->id)
+			return f->pa;
 	}
 	return SAMPLE_INVALID;
 }
 
 const char *format_id2paname(uint32_t id)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_formats); i++) {
-		if (id == audio_formats[i].id &&
-		    audio_formats[i].name != NULL)
-			return audio_formats[i].name;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_formats, f) {
+		if (id == f->id && f->name != NULL)
+			return f->name;
 	}
 	return "invalid";
 }
@@ -266,21 +262,18 @@ enum channel_position channel_id2pa(uint32_t id, uint32_t *aux)
 
 const char *channel_id2paname(uint32_t id, uint32_t *aux)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_channels); i++) {
-		if (id == audio_channels[i].channel &&
-		    audio_channels[i].name != NULL)
-			return audio_channels[i].name;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_channels, i) {
+		if (id == i->channel && i->name != NULL)
+			return i->name;
 	}
 	return audio_channels[CHANNEL_POSITION_AUX0 + ((*aux)++ & 31)].name;
 }
 
 uint32_t channel_paname2id(const char *name, size_t size)
 {
-	size_t i;
-	for (i = 0; i < SPA_N_ELEMENTS(audio_channels); i++) {
-		if (strncmp(name, audio_channels[i].name, size) == 0)
-			return audio_channels[i].channel;
+	SPA_FOR_EACH_ELEMENT_VAR(audio_channels, i) {
+		if (strncmp(name, i->name, size) == 0)
+			return i->channel;
 	}
 	return SPA_AUDIO_CHANNEL_UNKNOWN;
 }
@@ -389,7 +382,7 @@ struct encoding_info {
 
 static const struct encoding_info encoding_names[] = {
 	[ENCODING_ANY] = { "ANY", 0 },
-	[ENCODING_PCM] = { "PCM", 0 },
+	[ENCODING_PCM] = { "PCM", SPA_AUDIO_IEC958_CODEC_PCM },
 	[ENCODING_AC3_IEC61937] = { "AC3-IEC61937", SPA_AUDIO_IEC958_CODEC_AC3 },
 	[ENCODING_EAC3_IEC61937] = { "EAC3-IEC61937", SPA_AUDIO_IEC958_CODEC_EAC3 },
 	[ENCODING_MPEG_IEC61937] = { "MPEG-IEC61937", SPA_AUDIO_IEC958_CODEC_MPEG },
@@ -406,7 +399,7 @@ const char *format_encoding2name(enum encoding enc)
 		return encoding_names[enc].name;
 	return "INVALID";
 }
-static uint32_t format_encoding2id(enum encoding enc)
+uint32_t format_encoding2id(enum encoding enc)
 {
 	if (enc >= 0 && enc < (int)SPA_N_ELEMENTS(encoding_names) &&
 	    encoding_names[enc].name != NULL)
@@ -424,26 +417,8 @@ static enum encoding format_encoding_from_id(uint32_t id)
 	return ENCODING_ANY;
 }
 
-static inline int
-audio_raw_parse_opt(const struct spa_pod *format, struct spa_audio_info_raw *info)
-{
-	struct spa_pod *position = NULL;
-	int res;
-	info->flags = 0;
-	res = spa_pod_parse_object(format,
-			SPA_TYPE_OBJECT_Format, NULL,
-			SPA_FORMAT_AUDIO_format,        SPA_POD_OPT_Id(&info->format),
-			SPA_FORMAT_AUDIO_rate,          SPA_POD_OPT_Int(&info->rate),
-			SPA_FORMAT_AUDIO_channels,      SPA_POD_OPT_Int(&info->channels),
-			SPA_FORMAT_AUDIO_position,      SPA_POD_OPT_Pod(&position));
-	if (position == NULL ||
-	    !spa_pod_copy_array(position, SPA_TYPE_Id, info->position, SPA_AUDIO_MAX_CHANNELS))
-		SPA_FLAG_SET(info->flags, SPA_AUDIO_FLAG_UNPOSITIONED);
-
-	return res;
-}
-
-int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, struct channel_map *map,
+int format_parse_param(const struct spa_pod *param, bool collect,
+		struct sample_spec *ss, struct channel_map *map,
 		const struct sample_spec *def_ss, const struct channel_map *def_map)
 {
 	struct spa_audio_info info = { 0 };
@@ -457,28 +432,48 @@ int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, stru
 
 	switch (info.media_subtype) {
 	case SPA_MEDIA_SUBTYPE_raw:
+		if (spa_format_audio_raw_parse(param, &info.info.raw) < 0)
+	                return -ENOTSUP;
 		if (def_ss != NULL) {
 			if (ss != NULL)
 				*ss = *def_ss;
-			if (audio_raw_parse_opt(param, &info.info.raw) < 0)
-		                return -ENOTSUP;
 		} else {
-			if (spa_format_audio_raw_parse(param, &info.info.raw) < 0)
-		                return -ENOTSUP;
+			if (info.info.raw.format == 0 ||
+			    info.info.raw.rate == 0 ||
+			    info.info.raw.channels == 0 ||
+			    info.info.raw.channels > SPA_AUDIO_MAX_CHANNELS)
+				return -ENOTSUP;
 		}
 		break;
 	case SPA_MEDIA_SUBTYPE_iec958:
 	{
 		struct spa_audio_info_iec958 iec;
 
+		if (collect)
+			break;
+
 		if (spa_format_audio_iec958_parse(param, &iec) < 0)
 			return -ENOTSUP;
 
 		info.info.raw.format = SPA_AUDIO_FORMAT_S16;
-		info.info.raw.channels = 2;
 		info.info.raw.rate = iec.rate;
 		info.info.raw.position[0] = SPA_AUDIO_CHANNEL_FL;
 		info.info.raw.position[1] = SPA_AUDIO_CHANNEL_FR;
+		switch (iec.codec) {
+		case SPA_AUDIO_IEC958_CODEC_TRUEHD:
+		case SPA_AUDIO_IEC958_CODEC_DTSHD:
+			info.info.raw.channels = 8;
+			info.info.raw.position[2] = SPA_AUDIO_CHANNEL_FC;
+			info.info.raw.position[3] = SPA_AUDIO_CHANNEL_LFE;
+			info.info.raw.position[4] = SPA_AUDIO_CHANNEL_SL;
+			info.info.raw.position[5] = SPA_AUDIO_CHANNEL_SR;
+			info.info.raw.position[6] = SPA_AUDIO_CHANNEL_RL;
+			info.info.raw.position[7] = SPA_AUDIO_CHANNEL_RR;
+			break;
+		default:
+			info.info.raw.channels = 2;
+			break;
+		}
 		break;
 	}
 	default:
@@ -493,9 +488,11 @@ int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, stru
 		        ss->channels = info.info.raw.channels;
 	}
 	if (map) {
-		map->channels = info.info.raw.channels;
-		for (i = 0; i < map->channels; i++)
-			map->map[i] = info.info.raw.position[i];
+		if (info.info.raw.channels) {
+			map->channels = info.info.raw.channels;
+			for (i = 0; i < map->channels; i++)
+				map->map[i] = info.info.raw.position[i];
+		}
 	}
 	return 0;
 }
@@ -503,16 +500,53 @@ int format_parse_param(const struct spa_pod *param, struct sample_spec *ss, stru
 const struct spa_pod *format_build_param(struct spa_pod_builder *b, uint32_t id,
 		const struct sample_spec *spec, const struct channel_map *map)
 {
-	struct spa_audio_info_raw info;
+	struct spa_pod_frame f;
 
-	info = SPA_AUDIO_INFO_RAW_INIT(
-			.format = spec->format,
-			.channels = spec->channels,
-			.rate = spec->rate);
-	if (map)
-		channel_map_to_positions(map, info.position);
+	spa_pod_builder_push_object(b, &f, SPA_TYPE_OBJECT_Format, id);
+	spa_pod_builder_add(b,
+			SPA_FORMAT_mediaType,		SPA_POD_Id(SPA_MEDIA_TYPE_audio),
+			SPA_FORMAT_mediaSubtype,	SPA_POD_Id(SPA_MEDIA_SUBTYPE_raw),
+			0);
+	if (spec->format != SPA_AUDIO_FORMAT_UNKNOWN)
+		spa_pod_builder_add(b,
+			SPA_FORMAT_AUDIO_format,        SPA_POD_Id(spec->format), 0);
+	else {
+		spa_pod_builder_add(b,
+			SPA_FORMAT_AUDIO_format,	SPA_POD_CHOICE_ENUM_Id(14,
+								SPA_AUDIO_FORMAT_F32,
+								SPA_AUDIO_FORMAT_F32,
+								SPA_AUDIO_FORMAT_F32_OE,
+								SPA_AUDIO_FORMAT_S32,
+								SPA_AUDIO_FORMAT_S32_OE,
+								SPA_AUDIO_FORMAT_S24_32,
+								SPA_AUDIO_FORMAT_S24_32_OE,
+								SPA_AUDIO_FORMAT_S24,
+								SPA_AUDIO_FORMAT_S24_OE,
+								SPA_AUDIO_FORMAT_S16,
+								SPA_AUDIO_FORMAT_S16_OE,
+								SPA_AUDIO_FORMAT_ULAW,
+								SPA_AUDIO_FORMAT_ALAW,
+								SPA_AUDIO_FORMAT_U8),
+			0);
 
-	return spa_format_audio_raw_build(b, id, &info);
+	}
+
+        if (spec->rate != 0)
+		spa_pod_builder_add(b,
+			SPA_FORMAT_AUDIO_rate,          SPA_POD_Int(spec->rate), 0);
+	if (spec->channels != 0) {
+		spa_pod_builder_add(b,
+			SPA_FORMAT_AUDIO_channels,      SPA_POD_Int(spec->channels), 0);
+
+		if (map && map->channels == spec->channels) {
+			uint32_t positions[SPA_AUDIO_MAX_CHANNELS];
+			channel_map_to_positions(map, positions);
+                        spa_pod_builder_add(b, SPA_FORMAT_AUDIO_position,
+                                SPA_POD_Array(sizeof(uint32_t), SPA_TYPE_Id,
+                                        spec->channels, positions), 0);
+                }
+        }
+        return spa_pod_builder_pop(b, &f);
 }
 
 int format_info_from_spec(struct format_info *info, const struct sample_spec *ss,
@@ -576,7 +610,9 @@ static int add_int(struct format_info *info, const char *k, struct spa_pod *para
 		size_t size;
 		FILE *f;
 
-		f = open_memstream(&ptr, &size);
+		if ((f = open_memstream(&ptr, &size)) == NULL)
+			return -errno;
+
 		fprintf(f, "[");
 		for (i = 1; i < n_values; i++)
 			fprintf(f, "%s %d", i == 1 ? "" : ",", values[i]);
@@ -632,6 +668,7 @@ static int format_info_iec958_from_param(struct format_info *info, struct spa_po
 	default:
 		return -ENOTSUP;
 	}
+
 	if ((info->props = pw_properties_new(NULL, NULL)) == NULL)
 		return -errno;
 
