@@ -59,7 +59,7 @@ static const struct pw_impl_module_events module_events = {
 	.destroy = module_destroy
 };
 
-static int module_remap_sink_load(struct client *client, struct module *module)
+static int module_remap_sink_load(struct module *module)
 {
 	struct module_remap_sink_data *data = module->user_data;
 	FILE *f;
@@ -143,11 +143,11 @@ static void position_to_props(struct spa_audio_info_raw *info, struct pw_propert
 	pw_properties_set(props, SPA_KEY_AUDIO_POSITION, s);
 }
 
-struct module *create_module_remap_sink(struct impl *impl, const char *argument)
+static int module_remap_sink_prepare(struct module * const module)
 {
-	struct module *module;
-	struct module_remap_sink_data *d;
-	struct pw_properties *props = NULL, *playback_props = NULL, *capture_props = NULL;
+	struct module_remap_sink_data * const d = module->user_data;
+	struct pw_properties * const props = module->props;
+	struct pw_properties *playback_props = NULL, *capture_props = NULL;
 	const char *str, *master;
 	struct spa_audio_info_raw capture_info = { 0 };
 	struct spa_audio_info_raw playback_info = { 0 };
@@ -155,15 +155,12 @@ struct module *create_module_remap_sink(struct impl *impl, const char *argument)
 
 	PW_LOG_TOPIC_INIT(mod_topic);
 
-	props = pw_properties_new_dict(&SPA_DICT_INIT_ARRAY(module_remap_sink_info));
 	capture_props = pw_properties_new(NULL, NULL);
 	playback_props = pw_properties_new(NULL, NULL);
-	if (!props || !capture_props || !playback_props) {
+	if (!capture_props || !playback_props) {
 		res = -EINVAL;
 		goto out;
 	}
-	if (argument)
-		module_args_add_props(props, argument);
 
 	master = pw_properties_get(props, "master");
 	if (pw_properties_get(props, "sink_name") == NULL) {
@@ -204,7 +201,7 @@ struct module *create_module_remap_sink(struct impl *impl, const char *argument)
 		pw_properties_set(props, "master", NULL);
 	}
 
-	if (module_args_to_audioinfo(impl, props, &capture_info) < 0) {
+	if (module_args_to_audioinfo(module->impl, props, &capture_info) < 0) {
 		res = -EINVAL;
 		goto out;
 	}
@@ -235,30 +232,23 @@ struct module *create_module_remap_sink(struct impl *impl, const char *argument)
 	if (pw_properties_get(playback_props, PW_KEY_NODE_PASSIVE) == NULL)
 		pw_properties_set(playback_props, PW_KEY_NODE_PASSIVE, "true");
 
-	module = module_new(impl, sizeof(*d));
-	if (module == NULL) {
-		res = -errno;
-		goto out;
-	}
-
-	module->props = props;
-	d = module->user_data;
 	d->module = module;
 	d->capture_props = capture_props;
 	d->playback_props = playback_props;
 
-	return module;
+	return 0;
 out:
-	pw_properties_free(props);
 	pw_properties_free(playback_props);
 	pw_properties_free(capture_props);
-	errno = -res;
-	return NULL;
+
+	return res;
 }
 
 DEFINE_MODULE_INFO(module_remap_sink) = {
 	.name = "module-remap-sink",
-	.create = create_module_remap_sink,
+	.prepare = module_remap_sink_prepare,
 	.load = module_remap_sink_load,
 	.unload = module_remap_sink_unload,
+	.properties = &SPA_DICT_INIT_ARRAY(module_remap_sink_info),
+	.data_size = sizeof(struct module_remap_sink_data),
 };
