@@ -58,7 +58,7 @@ static const struct pw_impl_module_events module_events = {
 	.destroy = module_destroy
 };
 
-static int module_roc_sink_load(struct client *client, struct module *module)
+static int module_roc_sink_load(struct module *module)
 {
 	struct module_roc_sink_data *data = module->user_data;
 	FILE *f;
@@ -123,25 +123,22 @@ static const struct spa_dict_item module_roc_sink_info[] = {
 	{ PW_KEY_MODULE_VERSION, PACKAGE_VERSION },
 };
 
-struct module *create_module_roc_sink(struct impl *impl, const char *argument)
+static int module_roc_sink_prepare(struct module * const module)
 {
-	struct module *module;
-	struct module_roc_sink_data *d;
-	struct pw_properties *props = NULL, *sink_props = NULL, *roc_props = NULL;
+	struct module_roc_sink_data * const d = module->user_data;
+	struct pw_properties * const props = module->props;
+	struct pw_properties *sink_props = NULL, *roc_props = NULL;
 	const char *str;
 	int res;
 
 	PW_LOG_TOPIC_INIT(mod_topic);
 
-	props = pw_properties_new_dict(&SPA_DICT_INIT_ARRAY(module_roc_sink_info));
 	sink_props = pw_properties_new(NULL, NULL);
 	roc_props = pw_properties_new(NULL, NULL);
-	if (!props || !sink_props || !roc_props) {
+	if (!sink_props || !roc_props) {
 		res = -errno;
 		goto out;
 	}
-	if (argument != NULL)
-		module_args_add_props(props, argument);
 
 	if ((str = pw_properties_get(props, "sink_name")) != NULL) {
 		pw_properties_set(sink_props, PW_KEY_NODE_NAME, str);
@@ -185,30 +182,23 @@ struct module *create_module_roc_sink(struct impl *impl, const char *argument)
 		pw_properties_set(props, "fec_code", NULL);
 	}
 
-	module = module_new(impl, sizeof(*d));
-	if (module == NULL) {
-		res = -errno;
-		goto out;
-	}
-
-	module->props = props;
-	d = module->user_data;
 	d->module = module;
 	d->sink_props = sink_props;
 	d->roc_props = roc_props;
 
-	return module;
+	return 0;
 out:
-	pw_properties_free(props);
 	pw_properties_free(sink_props);
 	pw_properties_free(roc_props);
-	errno = -res;
-	return NULL;
+
+	return res;
 }
 
 DEFINE_MODULE_INFO(module_roc_sink) = {
 	.name = "module-roc-sink",
-	.create = create_module_roc_sink,
+	.prepare = module_roc_sink_prepare,
 	.load = module_roc_sink_load,
 	.unload = module_roc_sink_unload,
+	.properties = &SPA_DICT_INIT_ARRAY(module_roc_sink_info),
+	.data_size = sizeof(struct module_roc_sink_data),
 };
