@@ -28,13 +28,20 @@
 
 #include "../defs.h"
 #include "../module.h"
-#include "registry.h"
+
+#define NAME "zeroconf-discover"
+
+PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
+#define PW_LOG_TOPIC_DEFAULT mod_topic
+
 
 struct module_zeroconf_discover_data {
 	struct module *module;
 
 	struct spa_hook mod_listener;
 	struct pw_impl_module *mod;
+
+	uint32_t latency_msec;
 };
 
 static void module_destroy(void *data)
@@ -50,13 +57,28 @@ static const struct pw_impl_module_events module_events = {
 	.destroy = module_destroy
 };
 
-static int module_zeroconf_discover_load(struct client *client, struct module *module)
+static int module_zeroconf_discover_load(struct module *module)
 {
 	struct module_zeroconf_discover_data *data = module->user_data;
+	FILE *f;
+	char *args;
+	size_t size;
+
+	if ((f = open_memstream(&args, &size)) == NULL)
+		return -errno;
+
+	fprintf(f, "{");
+	if (data->latency_msec > 0)
+		fprintf(f, " pulse.latency = %u ", data->latency_msec);
+	fprintf(f, "}");
+	fclose(f);
 
 	data->mod = pw_context_load_module(module->impl->context,
 			"libpipewire-module-zeroconf-discover",
-			NULL, NULL);
+			args, NULL);
+
+	free(args);
+
 	if (data->mod == NULL)
 		return -errno;
 
@@ -67,7 +89,7 @@ static int module_zeroconf_discover_load(struct client *client, struct module *m
 	return 0;
 }
 
-static int module_zeroconf_discover_unload(struct client *client, struct module *module)
+static int module_zeroconf_discover_unload(struct module *module)
 {
 	struct module_zeroconf_discover_data *d = module->user_data;
 
@@ -80,47 +102,33 @@ static int module_zeroconf_discover_unload(struct client *client, struct module
 	return 0;
 }
 
-static const struct module_methods module_zeroconf_discover_methods = {
-	VERSION_MODULE_METHODS,
-	.load = module_zeroconf_discover_load,
-	.unload = module_zeroconf_discover_unload,
-};
-
 static const struct spa_dict_item module_zeroconf_discover_info[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Wim Taymans <wim.taymans@gmail.con>" },
 	{ PW_KEY_MODULE_DESCRIPTION, "mDNS/DNS-SD Service Discovery" },
-	{ PW_KEY_MODULE_USAGE, "" },
+	{ PW_KEY_MODULE_USAGE,
+		"latency_msec=<fixed latency in ms> " },
 	{ PW_KEY_MODULE_VERSION, PACKAGE_VERSION },
 };
 
-struct module *create_module_zeroconf_discover(struct impl *impl, const char *argument)
+static int module_zeroconf_discover_prepare(struct module * const module)
 {
-	struct module *module;
-	struct module_zeroconf_discover_data *d;
-	struct pw_properties *props = NULL;
-	int res;
+	PW_LOG_TOPIC_INIT(mod_topic);
 
-	props = pw_properties_new_dict(&SPA_DICT_INIT_ARRAY(module_zeroconf_discover_info));
-	if (props == NULL) {
-		res = -errno;
-		goto out;
-	}
-	if (argument != NULL)
-		module_args_add_props(props, argument);
+	struct pw_properties * const props = module->props;
+	struct module_zeroconf_discover_data * const data = module->user_data;
+	data->module = module;
 
-	module = module_new(impl, &module_zeroconf_discover_methods, sizeof(*d));
-	if (module == NULL) {
-		res = -errno;
-		goto out;
-	}
+	pw_properties_fetch_uint32(props, "latency_msec", &data->latency_msec);
 
-	module->props = props;
-	d = module->user_data;
-	d->module = module;
-
-	return module;
-out:
-	pw_properties_free(props);
-	errno = -res;
-	return NULL;
+	return 0;
 }
+
+DEFINE_MODULE_INFO(module_zeroconf_discover) = {
+	.name = "module-zeroconf-discover",
+	.load_once = true,
+	.prepare = module_zeroconf_discover_prepare,
+	.load = module_zeroconf_discover_load,
+	.unload = module_zeroconf_discover_unload,
+	.properties = &SPA_DICT_INIT_ARRAY(module_zeroconf_discover_info),
+	.data_size = sizeof(struct module_zeroconf_discover_data),
+};
