@@ -66,10 +66,63 @@
  *
  * Creates a new Sink to stream to an Airplay device.
  *
+ * Normally this sink is automatically created with \ref page_module_raop_discover
+ * with the right parameters but it is possible to manually create a RAOP sink
+ * as well.
+ *
  * ## Module Options
  *
+ * Options specific to the behavior of this module
+ *
+ * - `raop.hostname`: The hostname of the remote end.
+ * - `raop.port`: The port of the remote end.
+ * - `raop.transport`: The data transport to use, one of "udp" or "tcp". Defaults
+ *                    to "udp".
+ * - `raop.encryption.type`: The encryption type to use. One of "none", "RSA" or
+ *                    "auth_setup". Default is "none".
+ * - `raop.audio.codec`: The audio codec to use. Needs to be "PCM". Defaults to "PCM".
+ * - `raop.password`: The password to use.
+ * - `stream.props = {}`: properties to be passed to the sink stream
+ *
+ * Options with well-known behavior.
+ *
+ * - \ref PW_KEY_REMOTE_NAME
+ * - \ref PW_KEY_AUDIO_FORMAT
+ * - \ref PW_KEY_AUDIO_RATE
+ * - \ref PW_KEY_AUDIO_CHANNELS
+ * - \ref SPA_KEY_AUDIO_POSITION
+ * - \ref PW_KEY_NODE_NAME
+ * - \ref PW_KEY_NODE_DESCRIPTION
+ * - \ref PW_KEY_NODE_GROUP
+ * - \ref PW_KEY_NODE_LATENCY
+ * - \ref PW_KEY_NODE_VIRTUAL
+ * - \ref PW_KEY_MEDIA_CLASS
+ *
  * ## Example configuration
  *
+ *\code{.unparsed}
+ * context.modules = [
+ * {   name = libpipewire-module-raop-sink
+ *     args = {
+ *         # Set the remote address to tunnel to
+ *         raop.hostname = "my-raop-device"
+ *         raop.port = 8190
+ *         #raop.transport = "udp"
+ *         raop.encryption.type = "RSA"
+ *         #raop.audio.codec = "PCM"
+ *         #raop.password = "****"
+ *         #audio.format = "S16"
+ *         #audio.rate = 44100
+ *         #audio.channels = 2
+ *         #audio.position = [ FL FR ]
+ *         stream.props = {
+ *             # extra sink properties
+ *         }
+ *     }
+ * }
+ * ]
+ *\endcode
+ *
  * ## See also
  *
  * \ref page_module_raop_discover
@@ -104,9 +157,15 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 #define DEFAULT_CHANNELS 2
 #define DEFAULT_POSITION "[ FL FR ]"
 
-#define DEFAULT_LATENCY (DEFAULT_RATE*2)
+#define DEFAULT_LATENCY 22050
 
-#define MODULE_USAGE	"[ node.latency=<latency as fraction> ] "				\
+#define MODULE_USAGE	"[ raop.hostname=<name of host> ] "					\
+			"[ raop.port=<remote port> ] "						\
+			"[ raop.transport=<transport, default:udp> ] "				\
+			"[ raop.encryption.type=<encryption, default:none> ] "			\
+			"[ raop.audio.codec=PCM ] "						\
+			"[ raop.password=<password for auth> ] "				\
+			"[ node.latency=<latency as fraction> ] "				\
 			"[ node.name=<name of the nodes> ] "					\
 			"[ node.description=<description of the nodes> ] "			\
 			"[ audio.format=<format, default:"DEFAULT_FORMAT"> ] "			\
@@ -130,6 +189,7 @@ enum {
 enum {
 	CRYPTO_NONE,
 	CRYPTO_RSA,
+	CRYPTO_AUTH_SETUP,
 };
 enum {
 	CODEC_PCM,
@@ -257,7 +317,8 @@ static inline uint64_t ntp_now(int clockid)
 	return timespec_to_ntp(&now);
 }
 
-static int send_udp_sync_packet(struct impl *impl)
+static int send_udp_sync_packet(struct impl *impl,
+		struct sockaddr *dest_addr, socklen_t addrlen)
 {
 	uint32_t pkt[5];
 	uint32_t rtptime = impl->rtptime;
@@ -278,10 +339,11 @@ static int send_udp_sync_packet(struct impl *impl)
 	pw_log_debug("sync: delayed:%u now:%"PRIu64" rtptime:%u",
 			rtptime - delay, transmitted, rtptime);
 
-	return write(impl->control_fd, pkt, sizeof(pkt));
+	return sendto(impl->control_fd, pkt, sizeof(pkt), 0, dest_addr, addrlen);
 }
 
-static int send_udp_timing_packet(struct impl *impl, uint64_t remote, uint64_t received)
+static int send_udp_timing_packet(struct impl *impl, uint64_t remote, uint64_t received,
+		struct sockaddr *dest_addr, socklen_t addrlen)
 {
 	uint32_t pkt[8];
 	uint64_t transmitted;
@@ -299,7 +361,7 @@ static int send_udp_timing_packet(struct impl *impl, uint64_t remote, uint64_t r
 	pw_log_debug("sync: remote:%"PRIu64" received:%"PRIu64" transmitted:%"PRIu64,
 			remote, received, transmitted);
 
-	return write(impl->timing_fd, pkt, sizeof(pkt));
+	return sendto(impl->timing_fd, pkt, sizeof(pkt), 0, dest_addr, addrlen);
 }
 
 static int write_codec_pcm(void *dst, void *frames, uint32_t n_frames)
@@ -345,7 +407,7 @@ static int flush_to_udp_packet(struct impl *impl)
 	impl->sync++;
 	if (impl->first || impl->sync == impl->sync_period) {
 		impl->sync = 0;
-		send_udp_sync_packet(impl);
+		send_udp_sync_packet(impl, NULL, 0);
 	}
 	pkt[0] = htonl(0x80600000);
 	if (impl->first)
@@ -359,6 +421,7 @@ static int flush_to_udp_packet(struct impl *impl)
 
 	switch (impl->codec) {
 	case CODEC_PCM:
+	case CODEC_ALAC:
 		len = write_codec_pcm(dst, impl->buffer, n_frames);
 		break;
 	default:
@@ -366,14 +429,14 @@ static int flush_to_udp_packet(struct impl *impl)
 		memset(dst, 0, len);
 		break;
 	}
-	if (impl->encryption != CRYPTO_NONE)
+	if (impl->encryption == CRYPTO_RSA)
 		aes_encrypt(impl, dst, len);
 
 	impl->rtptime += n_frames;
 	impl->seq = (impl->seq + 1) & 0xffff;
 
 	pw_log_debug("send %u", len + 12);
-	res = write(impl->server_fd, pkt, len + 12);
+	res = send(impl->server_fd, pkt, len + 12, 0);
 
 	impl->first = false;
 
@@ -401,6 +464,7 @@ static int flush_to_tcp_packet(struct impl *impl)
 
 	switch (impl->codec) {
 	case CODEC_PCM:
+	case CODEC_ALAC:
 		len = write_codec_pcm(dst, impl->buffer, n_frames);
 		break;
 	default:
@@ -408,7 +472,7 @@ static int flush_to_tcp_packet(struct impl *impl)
 		memset(dst, 0, len);
 		break;
 	}
-	if (impl->encryption != CRYPTO_NONE)
+	if (impl->encryption == CRYPTO_RSA)
 		aes_encrypt(impl, dst, len);
 
 	pkt[0] |= htonl((uint32_t) len + 12);
@@ -417,7 +481,7 @@ static int flush_to_tcp_packet(struct impl *impl)
 	impl->seq = (impl->seq + 1) & 0xffff;
 
 	pw_log_debug("send %u", len + 16);
-	res = write(impl->server_fd, pkt, len + 16);
+	res = send(impl->server_fd, pkt, len + 16, 0);
 
 	impl->first = false;
 
@@ -515,10 +579,10 @@ static int create_udp_socket(struct impl *impl, uint16_t *port)
 
 		if (ip_version == 4) {
 			sa4.sin_port = htons(*port);
-			ret = bind(fd, &sa4, sizeof(sa4));
+			ret = bind(fd, (struct sockaddr*)&sa4, sizeof(sa4));
 		} else {
 			sa6.sin6_port = htons(*port);
-			ret = bind(fd, &sa6, sizeof(sa6));
+			ret = bind(fd, (struct sockaddr*)&sa6, sizeof(sa6));
 		}
 		if (ret == 0)
 			break;
@@ -593,9 +657,12 @@ on_timing_source_io(void *data, int fd, uint32_t mask)
 
 	if (mask & SPA_IO_IN) {
 		uint64_t remote, received;
+		struct sockaddr_storage sender;
+		socklen_t sender_size = sizeof(sender);
 
 		received = ntp_now(CLOCK_MONOTONIC);
-		bytes = read(impl->timing_fd, packet, sizeof(packet));
+		bytes = recvfrom(impl->timing_fd, packet, sizeof(packet), 0,
+				(struct sockaddr*)&sender, &sender_size);
 		if (bytes < 0) {
 			pw_log_debug("error reading timing packet: %m");
 			return;
@@ -609,7 +676,11 @@ on_timing_source_io(void *data, int fd, uint32_t mask)
 			return;
 
 		remote = ((uint64_t)ntohl(packet[6])) << 32 | ntohl(packet[7]);
-		send_udp_timing_packet(impl, remote, received);
+		if (send_udp_timing_packet(impl, remote, received,
+				(struct sockaddr *)&sender, sender_size) < 0) {
+			pw_log_warn("error sending timing packet");
+			return;
+		}
 	}
 }
 
@@ -653,9 +724,10 @@ on_control_source_io(void *data, int fd, uint32_t mask)
 	}
 }
 
-static void rtsp_flush_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_flush_reply(void *data, int status, const struct spa_dict *headers)
 {
 	pw_log_info("reply %d", status);
+	return 0;
 }
 
 static int rtsp_do_flush(struct impl *impl)
@@ -680,7 +752,7 @@ static int rtsp_do_flush(struct impl *impl)
 	return res;
 }
 
-static void rtsp_record_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_record_reply(void *data, int status, const struct spa_dict *headers)
 {
 	struct impl *impl = data;
 	const char *str;
@@ -689,6 +761,7 @@ static void rtsp_record_reply(void *data, int status, const struct spa_dict *hea
 	uint8_t buffer[1024];
 	struct spa_pod_builder b;
 	struct spa_latency_info latency;
+	char progress[128];
 
 	pw_log_info("reply %d", status);
 
@@ -713,6 +786,10 @@ static void rtsp_record_reply(void *data, int status, const struct spa_dict *hea
 	impl->sync = 0;
 	impl->sync_period = impl->info.rate / (impl->block_size / impl->frame_size);
 	impl->recording = true;
+
+	snprintf(progress, sizeof(progress), "progress: %s/%s/%s\r\n", "0", "0", "0");
+	return pw_rtsp_client_send(impl->rtsp, "SET_PARAMETER", NULL,
+			"text/parameters", progress, NULL, NULL);
 }
 
 static int rtsp_do_record(struct impl *impl)
@@ -766,7 +843,7 @@ error:
 	pw_loop_update_io(impl->loop, impl->server_source, 0);
 }
 
-static void rtsp_setup_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_setup_reply(void *data, int status, const struct spa_dict *headers)
 {
 	struct impl *impl = data;
 	const char *str, *state = NULL, *s;
@@ -778,13 +855,13 @@ static void rtsp_setup_reply(void *data, int status, const struct spa_dict *head
 
 	if ((str = spa_dict_lookup(headers, "Session")) == NULL) {
 		pw_log_error("missing Session header");
-		return;
+		return 0;
 	}
 	pw_properties_set(impl->headers, "Session", str);
 
 	if ((str = spa_dict_lookup(headers, "Transport")) == NULL) {
 		pw_log_error("missing Transport header");
-		return;
+		return 0;
 	}
 
 	impl->server_port = control_port = timing_port = 0;
@@ -801,18 +878,21 @@ static void rtsp_setup_reply(void *data, int status, const struct spa_dict *head
 	}
 	if (impl->server_port == 0) {
 		pw_log_error("missing server port in Transport");
-		return;
+		return 0;
 	}
 
-	pw_getrandom(&impl->seq, sizeof(impl->seq), 0);
-	pw_getrandom(&impl->rtptime, sizeof(impl->rtptime), 0);
+	if (pw_getrandom(&impl->seq, sizeof(impl->seq), 0) < 0 ||
+	    pw_getrandom(&impl->rtptime, sizeof(impl->rtptime), 0) <  0) {
+		pw_log_error("error generating random seq and rtptime: %m");
+		return 0;
+	}
 
 	pw_log_info("server port:%u", impl->server_port);
 
 	switch (impl->protocol) {
 	case PROTO_TCP:
-		if ((impl->server_fd = connect_socket(impl, SOCK_STREAM, -1, impl->server_port)) <= 0)
-			return;
+		if ((impl->server_fd = connect_socket(impl, SOCK_STREAM, -1, impl->server_port)) < 0)
+			return impl->server_fd;
 
 		impl->server_source = pw_loop_add_io(impl->loop, impl->server_fd,
 				SPA_IO_OUT, false, on_server_source_io, impl);
@@ -821,22 +901,20 @@ static void rtsp_setup_reply(void *data, int status, const struct spa_dict *head
 	case PROTO_UDP:
 		if (control_port == 0 || timing_port == 0) {
 			pw_log_error("missing UDP ports in Transport");
-			return;
+			return 0;
 		}
 		pw_log_info("control:%u timing:%u", control_port, timing_port);
 
-		if ((impl->server_fd = connect_socket(impl, SOCK_DGRAM, -1, impl->server_port)) <= 0)
-			return;
-		if ((impl->control_fd = connect_socket(impl, SOCK_DGRAM, impl->control_fd, control_port)) <= 0)
-			return;
-		if ((impl->timing_fd = connect_socket(impl, SOCK_DGRAM, impl->timing_fd, timing_port)) <= 0)
-			return;
+		if ((impl->server_fd = connect_socket(impl, SOCK_DGRAM, -1, impl->server_port)) < 0)
+			return impl->server_fd;
+		if ((impl->control_fd = connect_socket(impl, SOCK_DGRAM, impl->control_fd, control_port)) < 0)
+			return impl->control_fd;
+		if ((impl->timing_fd = connect_socket(impl, SOCK_DGRAM, impl->timing_fd, timing_port)) < 0)
+			return impl->timing_fd;
 
 		ntp = ntp_now(CLOCK_MONOTONIC);
-		send_udp_timing_packet(impl, ntp, ntp);
+		send_udp_timing_packet(impl, ntp, ntp, NULL, 0);
 
-		impl->timing_source = pw_loop_add_io(impl->loop, impl->timing_fd,
-				SPA_IO_IN, false, on_timing_source_io, impl);
 		impl->control_source = pw_loop_add_io(impl->loop, impl->control_fd,
 				SPA_IO_IN, false, on_control_source_io, impl);
 
@@ -845,8 +923,9 @@ static void rtsp_setup_reply(void *data, int status, const struct spa_dict *head
 			rtsp_do_record(impl);
 		break;
 	default:
-		return;
+		return 0;
 	}
+	return 0;
 }
 
 static int rtsp_do_setup(struct impl *impl)
@@ -868,6 +947,9 @@ static int rtsp_do_setup(struct impl *impl)
 		if (impl->control_fd < 0 || impl->timing_fd < 0)
 			goto error;
 
+		impl->timing_source = pw_loop_add_io(impl->loop, impl->timing_fd,
+				SPA_IO_IN, false, on_timing_source_io, impl);
+
 		pw_properties_setf(impl->headers, "Transport",
 				"RTP/AVP/UDP;unicast;interleaved=0-1;mode=record;"
 				"control_port=%u;timing_port=%u",
@@ -894,7 +976,7 @@ error:
 	return -EIO;
 }
 
-static void rtsp_announce_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_announce_reply(void *data, int status, const struct spa_dict *headers)
 {
 	struct impl *impl = data;
 
@@ -902,7 +984,7 @@ static void rtsp_announce_reply(void *data, int status, const struct spa_dict *h
 
 	pw_properties_set(impl->headers, "Apple-Challenge", NULL);
 
-	rtsp_do_setup(impl);
+	return rtsp_do_setup(impl);
 }
 
 static void base64_encode(const uint8_t *data, size_t len, char *enc, char pad)
@@ -988,8 +1070,9 @@ static int rtsp_do_announce(struct impl *impl)
 	char iv[16*2];
 	int res, frames, i, ip_version;
 	char *sdp;
-        char local_ip[256];
-
+	char local_ip[256];
+	int min_latency;
+	min_latency = DEFAULT_LATENCY;
 	host = pw_properties_get(impl->props, "raop.hostname");
 
 	if (impl->protocol == PROTO_TCP)
@@ -1016,10 +1099,26 @@ static int rtsp_do_announce(struct impl *impl)
 				ip_version, host, frames);
 		break;
 
+	case CRYPTO_AUTH_SETUP:
+		asprintf(&sdp, "v=0\r\n"
+				"o=iTunes %s 0 IN IP%d %s\r\n"
+				"s=iTunes\r\n"
+				"c=IN IP%d %s\r\n"
+				"t=0 0\r\n"
+				"m=audio 0 RTP/AVP 96\r\n"
+				"a=rtpmap:96 AppleLossless\r\n"
+				"a=fmtp:96 %d 0 16 40 10 14 2 255 0 0 44100\r\n"
+				"a=min-latency:%d",
+				impl->session_id, ip_version, local_ip,
+				ip_version, host, frames, min_latency);
+		break;
+
 	case CRYPTO_RSA:
-		pw_getrandom(impl->key, sizeof(impl->key), 0);
+		if (pw_getrandom(impl->key, sizeof(impl->key), 0) < 0 ||
+		    pw_getrandom(impl->iv, sizeof(impl->iv), 0) < 0)
+			return -errno;
+
 		AES_set_encrypt_key(impl->key, 128, &impl->aes);
-		pw_getrandom(impl->iv, sizeof(impl->iv), 0);
 
 		i = rsa_encrypt(impl->key, 16, rsakey);
 	        base64_encode(rsakey, i, key, '=');
@@ -1048,6 +1147,27 @@ static int rtsp_do_announce(struct impl *impl)
 	return res;
 }
 
+static int rtsp_auth_setup_reply(void *data, int status, const struct spa_dict *headers)
+{
+	struct impl *impl = data;
+
+	pw_log_info("reply %d", status);
+
+	return rtsp_do_announce(impl);
+}
+
+static int rtsp_do_auth_setup(struct impl *impl)
+{
+	static const unsigned char content[33] =
+		"\x01"
+		"\x59\x02\xed\xe9\x0d\x4e\xf2\xbd\x4c\xb6\x8a\x63\x30\x03\x82\x07"
+		"\xa9\x4d\xbd\x50\xd8\xaa\x46\x5b\x5d\x8c\x01\x2a\x0c\x7e\x1d\x4e";
+
+	return pw_rtsp_client_url_send(impl->rtsp, "/auth-setup", "POST", &impl->headers->dict,
+				       "application/octet-stream", content, sizeof(content),
+				       rtsp_auth_setup_reply, impl);
+}
+
 static const char *find_attr(char **tokens, const char *key)
 {
 	int i;
@@ -1066,16 +1186,22 @@ static const char *find_attr(char **tokens, const char *key)
 	return NULL;
 }
 
-static void rtsp_auth_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_auth_reply(void *data, int status, const struct spa_dict *headers)
 {
 	struct impl *impl = data;
+	int res = 0;
+
 	pw_log_info("auth %d", status);
 
 	switch (status) {
 	case 200:
-		rtsp_do_announce(impl);
+		if (impl->encryption == CRYPTO_AUTH_SETUP)
+			res = rtsp_do_auth_setup(impl);
+		else
+			res = rtsp_do_announce(impl);
 		break;
 	}
+	return res;
 }
 
 SPA_PRINTF_FUNC(2,3)
@@ -1124,7 +1250,7 @@ static int rtsp_do_auth(struct impl *impl, const struct spa_dict *headers)
 		spa_scnprintf(auth, sizeof(auth), "Basic %s", enc);
 	}
 	else if (spa_streq(tokens[0], "Digest")) {
-		const char *realm, *nonce;
+		const char *realm, *nonce, *url;
 		char h1[MD5_HASH_LENGTH+1];
 		char h2[MD5_HASH_LENGTH+1];
 		char resp[MD5_HASH_LENGTH+1];
@@ -1134,13 +1260,15 @@ static int rtsp_do_auth(struct impl *impl, const struct spa_dict *headers)
 		if (realm == NULL || nonce == NULL)
 			goto error;
 
+		url = pw_rtsp_client_get_url(impl->rtsp);
+
 		MD5_hash(h1, "%s:%s:%s", DEFAULT_USER_NAME, realm, impl->password);
-		MD5_hash(h2, "OPTIONS:*");
+		MD5_hash(h2, "OPTIONS:%s", url);
 		MD5_hash(resp, "%s:%s:%s", h1, nonce, h2);
 
 		spa_scnprintf(auth, sizeof(auth),
-				"Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"*\", response=\"%s\"",
-				DEFAULT_USER_NAME, realm, nonce, resp);
+				"username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", response=\"%s\"",
+				DEFAULT_USER_NAME, realm, nonce, url, resp);
 	}
 	else
 		goto error;
@@ -1158,19 +1286,25 @@ error:
 	return -EINVAL;
 }
 
-static void rtsp_options_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_options_reply(void *data, int status, const struct spa_dict *headers)
 {
 	struct impl *impl = data;
+	int res = 0;
+
 	pw_log_info("options %d", status);
 
 	switch (status) {
 	case 401:
-		rtsp_do_auth(impl, headers);
+		res = rtsp_do_auth(impl, headers);
 		break;
 	case 200:
-		rtsp_do_announce(impl);
+		if (impl->encryption == CRYPTO_AUTH_SETUP)
+			res = rtsp_do_auth_setup(impl);
+		else
+			res = rtsp_do_announce(impl);
 		break;
 	}
+	return res;
 }
 
 static void rtsp_connected(void *data)
@@ -1184,11 +1318,15 @@ static void rtsp_connected(void *data)
 
 	impl->connected = true;
 
-	pw_getrandom(sci, sizeof(sci), 0);
+	if (pw_getrandom(sci, sizeof(sci), 0) < 0 ||
+	    pw_getrandom(rac, sizeof(rac), 0) < 0) {
+		pw_log_error("error generating random data: %m");
+		return;
+	}
+
 	pw_properties_setf(impl->headers, "Client-Instance",
 			"%08x%08x", sci[0], sci[1]);
 
-	pw_getrandom(rac, sizeof(rac), 0);
 	base64_encode(rac, sizeof(rac), sac, '\0');
 	pw_properties_set(impl->headers, "Apple-Challenge", sac);
 
@@ -1201,29 +1339,29 @@ static void rtsp_connected(void *data)
 static void connection_cleanup(struct impl *impl)
 {
 	impl->ready = false;
-	if (impl->server_fd != -1) {
-		close(impl->server_fd);
-		impl->server_fd = -1;
-	}
-	if (impl->control_fd != -1) {
-		close(impl->control_fd);
-		impl->control_fd = -1;
-	}
-	if (impl->timing_fd != -1) {
-		close(impl->timing_fd);
-		impl->timing_fd = -1;
-	}
 	if (impl->server_source != NULL) {
 		pw_loop_destroy_source(impl->loop, impl->server_source);
 		impl->server_source = NULL;
 	}
+	if (impl->server_fd >= 0) {
+		close(impl->server_fd);
+		impl->server_fd = -1;
+	}
+	if (impl->control_source != NULL) {
+		pw_loop_destroy_source(impl->loop, impl->control_source);
+		impl->control_source = NULL;
+	}
+	if (impl->control_fd >= 0) {
+		close(impl->control_fd);
+		impl->control_fd = -1;
+	}
 	if (impl->timing_source != NULL) {
 		pw_loop_destroy_source(impl->loop, impl->timing_source);
 		impl->timing_source = NULL;
 	}
-	if (impl->control_source != NULL) {
-		pw_loop_destroy_source(impl->loop, impl->control_source);
-		impl->control_source = NULL;
+	if (impl->timing_fd >= 0) {
+		close(impl->timing_fd);
+		impl->timing_fd = -1;
 	}
 }
 
@@ -1294,13 +1432,15 @@ static int rtsp_do_connect(struct impl *impl)
 	if (hostname == NULL || port == NULL)
 		return -EINVAL;
 
-	pw_getrandom(&session_id, sizeof(session_id), 0);
+	if (pw_getrandom(&session_id, sizeof(session_id), 0) < 0)
+		return -errno;
+
 	spa_scnprintf(impl->session_id, sizeof(impl->session_id), "%u", session_id);
 
 	return pw_rtsp_client_connect(impl->rtsp, hostname, atoi(port), impl->session_id);
 }
 
-static void rtsp_teardown_reply(void *data, int status, const struct spa_dict *headers)
+static int rtsp_teardown_reply(void *data, int status, const struct spa_dict *headers)
 {
 	struct impl *impl = data;
 	const char *str;
@@ -1313,6 +1453,7 @@ static void rtsp_teardown_reply(void *data, int status, const struct spa_dict *h
 		if (spa_streq(str, "close"))
 			pw_rtsp_client_disconnect(impl->rtsp);
 	}
+	return 0;
 }
 
 static int rtsp_do_teardown(struct impl *impl)
@@ -1648,6 +1789,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		impl->encryption = CRYPTO_NONE;
 	else if (spa_streq(str, "RSA"))
 		impl->encryption = CRYPTO_RSA;
+	else if (spa_streq(str, "auth_setup"))
+		impl->encryption = CRYPTO_AUTH_SETUP;
 	else {
 		pw_log_error( "can't handle encryption type %s", str);
 		res = -EINVAL;
@@ -1658,6 +1801,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		str = "PCM";
 	if (spa_streq(str, "PCM"))
 		impl->codec = CODEC_PCM;
+	else if (spa_streq(str, "ALAC"))
+		impl->codec = CODEC_ALAC;
 	else {
 		pw_log_error( "can't handle codec type %s", str);
 		res = -EINVAL;
