@@ -48,6 +48,26 @@
 #include "module-zeroconf-discover/avahi-poll.h"
 
 /** \page page_module_zeroconf_discover PipeWire Module: Zeroconf Discover
+ *
+ * Use zeroconf to detect and load module-pulse-tunnel with the right
+ * parameters. This will automatically create sinks and sources to stream
+ * audio to/from remote PulseAudio servers. It also works with
+ * module-protocol-pulse.
+ *
+ * ## Module Options
+ *
+ * - `pulse.latency`: the latency to end-to-end latency in milliseconds to
+ *                    maintain (Default 200ms).
+ *
+ * ## Example configuration
+ *
+ *\code{.unparsed}
+ * context.modules = [
+ * {   name = libpipewire-module-zeroconf-discover
+ *     args = { }
+ * }
+ * ]
+ *\endcode
  */
 
 #define NAME "zeroconf-discover"
@@ -55,7 +75,7 @@
 PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 #define PW_LOG_TOPIC_DEFAULT mod_topic
 
-#define MODULE_USAGE	" "
+#define MODULE_USAGE	"pulse.latency=<latency in msec> "
 
 static const struct spa_dict_item module_props[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Wim Taymans <wim.taymans@gmail.com>" },
@@ -73,7 +93,6 @@ struct impl {
 
 	struct pw_impl_module *module;
 	struct spa_hook module_listener;
-	struct pw_work_queue *work;
 
 	struct pw_properties *properties;
 
@@ -83,8 +102,6 @@ struct impl {
 	AvahiServiceBrowser *source_browser;
 
 	struct spa_list tunnel_list;
-
-	unsigned int unloading:1;
 };
 
 struct tunnel_info {
@@ -95,7 +112,7 @@ struct tunnel_info {
 	const char *domain;
 };
 
-#define TUNNEL_INFO(...) (struct tunnel_info){ __VA_ARGS__ }
+#define TUNNEL_INFO(...) ((struct tunnel_info){ __VA_ARGS__ })
 
 struct tunnel {
 	struct spa_list link;
@@ -106,20 +123,6 @@ struct tunnel {
 
 static int start_client(struct impl *impl);
 
-static void do_unload_module(void *obj, void *data, int res, uint32_t id)
-{
-	struct impl *impl = data;
-	pw_impl_module_destroy(impl->module);
-}
-
-static void unload_module(struct impl *impl)
-{
-	if (!impl->unloading) {
-		impl->unloading = true;
-		pw_work_queue_add(impl->work, impl, 0, do_unload_module, impl);
-	}
-}
-
 static struct tunnel *make_tunnel(struct impl *impl, const struct tunnel_info *info)
 {
 	struct tunnel *t;
@@ -173,8 +176,6 @@ static void impl_free(struct impl *impl)
 	if (impl->avahi_poll)
 		pw_avahi_poll_free(impl->avahi_poll);
 	pw_properties_free(impl->properties);
-	if (impl->work)
-		pw_work_queue_cancel(impl->work, impl, SPA_ID_INVALID);
 	free(impl);
 }
 
@@ -190,15 +191,17 @@ static const struct pw_impl_module_events module_events = {
 	.destroy = module_destroy,
 };
 
-static void pw_properties_from_avahi_string(const char *key, const char *value, struct pw_properties *props) {
+static void pw_properties_from_avahi_string(const char *key, const char *value,
+		struct pw_properties *props)
+{
 	if (spa_streq(key, "device")) {
 		pw_properties_set(props, PW_KEY_NODE_TARGET, value);
 	}
 	else if (spa_streq(key, "rate")) {
-		pw_properties_setf(props, PW_KEY_AUDIO_RATE, "%u", atoi(value));
+		pw_properties_set(props, PW_KEY_AUDIO_RATE, value);
 	}
 	else if (spa_streq(key, "channels")) {
-		pw_properties_setf(props, PW_KEY_AUDIO_CHANNELS, "%u", atoi(value));
+		pw_properties_set(props, PW_KEY_AUDIO_CHANNELS, value);
 	}
 	else if (spa_streq(key, "channel_map")) {
 		struct channel_map channel_map;
@@ -209,16 +212,18 @@ static void pw_properties_from_avahi_string(const char *key, const char *value,
 		channel_map_parse(value, &channel_map);
 		channel_map_to_positions(&channel_map, pos);
 
-		p = s = alloca(4 + channel_map.channels * 6);
-		p += snprintf(p, 6, "[");
+		p = s = alloca(4 + channel_map.channels * 8);
+		p += spa_scnprintf(p, 2, "[");
 		for (i = 0; i < channel_map.channels; i++)
-			p += snprintf(p, 6, "%s%s", i == 0 ? "" : ",",
+			p += spa_scnprintf(p, 8, "%s%s", i == 0 ? "" : ",",
 				channel_id2name(pos[i]));
-		p += snprintf(p, 6, "]");
+		p += spa_scnprintf(p, 2, "]");
 		pw_properties_set(props, SPA_KEY_AUDIO_POSITION, s);
 	}
 	else if (spa_streq(key, "format")) {
-		pw_properties_set(props, PW_KEY_AUDIO_FORMAT, value);
+		uint32_t fmt = format_paname2id(value, strlen(value));
+		if (fmt != SPA_AUDIO_FORMAT_UNKNOWN)
+			pw_properties_set(props, PW_KEY_AUDIO_FORMAT, format_id2name(fmt));
 	}
 	else if (spa_streq(key, "icon-name")) {
 		pw_properties_set(props, PW_KEY_DEVICE_ICON_NAME, value);
@@ -309,7 +314,7 @@ static void resolver_cb(AvahiServiceResolver *r, AvahiIfIndex interface, AvahiPr
 		pw_properties_setf(props, PW_KEY_NODE_NAME,
 				"tunnel.%s", host_name);
 
-	str = strstr(type, "sink") ? "playback" : "capture";
+	str = strstr(type, "sink") ? "sink" : "source";
 	pw_properties_set(props, "tunnel.mode", str);
 
 	if (a->proto == AVAHI_PROTO_INET6 &&
@@ -346,7 +351,14 @@ static void resolver_cb(AvahiServiceResolver *r, AvahiIfIndex interface, AvahiPr
 				_("%s on %s"), desc, fqdn);
 	}
 
-	f = open_memstream(&args, &size);
+	if ((str = pw_properties_get(impl->properties, "pulse.latency")) != NULL)
+		pw_properties_set(props, "pulse.latency", str);
+
+	if ((f = open_memstream(&args, &size)) == NULL) {
+		pw_log_error("Can't open memstream: %m");
+		goto done;
+	}
+
 	fprintf(f, "{");
 	pw_properties_serialize_dict(f, &props->dict, 0);
 	fprintf(f, " stream.props = {");
@@ -481,7 +493,7 @@ static void client_callback(AvahiClient *c, AvahiClientState state, void *userda
 	}
 	return;
 error:
-	unload_module(impl);
+	pw_impl_module_schedule_destroy(impl->module);
 }
 
 static int start_client(struct impl *impl)
@@ -492,7 +504,7 @@ static int start_client(struct impl *impl)
 					client_callback, impl,
 					&res)) == NULL) {
 		pw_log_error("can't create client: %s", avahi_strerror(res));
-		unload_module(impl);
+		pw_impl_module_schedule_destroy(impl->module);
 		return -EIO;
 	}
 	return 0;
@@ -537,10 +549,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->context = context;
 	impl->properties = props;
 
-	impl->work = pw_context_get_work_queue(context);
-	if (impl->work == NULL)
-		goto error_errno;
-
 	pw_impl_module_add_listener(module, &impl->module_listener, &module_events, impl);
 
 	pw_impl_module_update_properties(module, &SPA_DICT_INIT_ARRAY(module_props));
