@@ -34,6 +34,7 @@
 #include <spa/utils/string.h>
 #include <spa/debug/types.h>
 #include <spa/pod/filter.h>
+#include <spa/pod/dynamic.h>
 
 #include "pipewire/impl.h"
 #include "pipewire/private.h"
@@ -208,44 +209,63 @@ SPA_EXPORT
 int pw_impl_port_init_mix(struct pw_impl_port *port, struct pw_impl_port_mix *mix)
 {
 	uint32_t port_id;
+	struct pw_impl_node *node = port->node;
 	int res = 0;
 
 	port_id = pw_map_insert_new(&port->mix_port_map, mix);
 	if (port_id == SPA_ID_INVALID)
 		return -errno;
 
+	if ((res = spa_node_add_port(port->mix, port->direction, port_id, NULL)) < 0 &&
+	    res != -ENOTSUP)
+		goto error_remove_map;
+
 	mix->port.direction = port->direction;
 	mix->port.port_id = port_id;
-
-	spa_list_append(&port->mix_list, &mix->link);
-	port->n_mix++;
 	mix->p = port;
 
-	spa_node_add_port(port->mix, port->direction, port_id, NULL);
-
-	res = pw_impl_port_call_init_mix(port, mix);
+	if ((res = pw_impl_port_call_init_mix(port, mix)) < 0)
+		goto error_remove_port;
 
 	/* set the same format on the mixer as on the port if any */
 	{
 		uint32_t idx = 0;
 		uint8_t buffer[1024];
-		struct spa_pod_builder b;
+		struct spa_pod_dynamic_builder b;
 		struct spa_pod *param;
 
-		spa_pod_builder_init(&b, buffer, sizeof(buffer));
+		spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+
 		if (spa_node_port_enum_params_sync(port->mix,
 				pw_direction_reverse(port->direction), 0,
-				SPA_PARAM_Format, &idx, NULL, &param, &b) == 1) {
+				SPA_PARAM_Format, &idx, NULL, &param, &b.b) == 1) {
 			spa_node_port_set_param(port->mix,
 				port->direction, port_id,
 				SPA_PARAM_Format, 0, param);
 		}
+		spa_pod_dynamic_builder_clean(&b);
 	}
 
+	spa_list_append(&port->mix_list, &mix->link);
+	port->n_mix++;
+
 	pw_log_debug("%p: init mix n_mix:%d %d.%d io:%p: (%s)", port,
 			port->n_mix, port->port_id, mix->port.port_id,
 			mix->io, spa_strerror(res));
 
+	if (port->n_mix == 1) {
+		pw_log_debug("%p: setting port io", port);
+		spa_node_port_set_io(node->node,
+				     port->direction, port->port_id,
+				     SPA_IO_Buffers,
+				     &port->rt.io, sizeof(port->rt.io));
+	}
+	return res;
+
+error_remove_port:
+	spa_node_remove_port(port->mix, port->direction, port_id);
+error_remove_map:
+	pw_map_remove(&port->mix_port_map, port_id);
 	return res;
 }
 
@@ -254,6 +274,7 @@ int pw_impl_port_release_mix(struct pw_impl_port *port, struct pw_impl_port_mix
 {
 	int res = 0;
 	uint32_t port_id = mix->port.port_id;
+	struct pw_impl_node *node = port->node;
 
 	pw_map_remove(&port->mix_port_map, port_id);
 	spa_list_remove(&mix->link);
@@ -261,11 +282,20 @@ int pw_impl_port_release_mix(struct pw_impl_port *port, struct pw_impl_port_mix
 
 	res = pw_impl_port_call_release_mix(port, mix);
 
-	spa_node_remove_port(port->mix, port->direction, port_id);
+	if ((res = spa_node_remove_port(port->mix, port->direction, port_id)) < 0 &&
+	    res != -ENOTSUP)
+		pw_log_warn("can't remove mix port %d: %s", port_id, spa_strerror(res));
 
 	pw_log_debug("%p: release mix %d %d.%d", port,
 			port->n_mix, port->port_id, mix->port.port_id);
 
+	if (port->n_mix == 0) {
+		pw_log_debug("%p: clearing port io", port);
+		spa_node_port_set_io(node->node,
+				     port->direction, port->port_id,
+				     SPA_IO_Buffers,
+				     NULL, sizeof(port->rt.io));
+	}
 	return res;
 }
 
@@ -583,8 +613,10 @@ static int setup_mixer(struct pw_impl_port *port, const struct spa_pod *param)
 	int res;
 	const char *fallback_lib, *factory_name;
 	struct spa_handle *handle;
-	struct spa_dict_item items[1];
+	struct spa_dict_item items[2];
+	char quantum_limit[16];
 	void *iface;
+	struct pw_context *context = port->node->context;
 
 	if ((res = spa_format_parse(param, &media_type, &media_subtype)) < 0)
 		return res;
@@ -634,7 +666,11 @@ static int setup_mixer(struct pw_impl_port *port, const struct spa_pod *param)
 	}
 
 	items[0] = SPA_DICT_ITEM_INIT(SPA_KEY_LIBRARY_NAME, fallback_lib);
-	handle = pw_context_load_spa_handle(port->node->context, factory_name,
+	spa_scnprintf(quantum_limit, sizeof(quantum_limit), "%u",
+			context->settings.clock_quantum_limit);
+	items[1] = SPA_DICT_ITEM_INIT("clock.quantum-limit", quantum_limit);
+
+	handle = pw_context_load_spa_handle(context, factory_name,
 			&SPA_DICT_INIT_ARRAY(items));
 	if (handle == NULL)
 		return -errno;
@@ -827,10 +863,10 @@ static const struct pw_resource_events resource_events = {
 };
 
 static int
-global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
+global_bind(void *object, struct pw_impl_client *client, uint32_t permissions,
 	       uint32_t version, uint32_t id)
 {
-	struct pw_impl_port *this = _data;
+	struct pw_impl_port *this = object;
 	struct pw_global *global = this->global;
 	struct pw_resource *resource;
 	struct resource_data *data;
@@ -866,9 +902,9 @@ error_resource:
 	return res;
 }
 
-static void global_destroy(void *object)
+static void global_destroy(void *data)
 {
-	struct pw_impl_port *port = object;
+	struct pw_impl_port *port = data;
 	spa_hook_remove(&port->global_listener);
 	port->global = NULL;
 	pw_impl_port_destroy(port);
@@ -883,6 +919,7 @@ int pw_impl_port_register(struct pw_impl_port *port,
 		     struct pw_properties *properties)
 {
 	static const char * const keys[] = {
+		PW_KEY_OBJECT_SERIAL,
 		PW_KEY_OBJECT_PATH,
 		PW_KEY_FORMAT_DSP,
 		PW_KEY_NODE_ID,
@@ -918,6 +955,8 @@ int pw_impl_port_register(struct pw_impl_port *port,
 	port->info.id = port->global->id;
 	pw_properties_setf(port->properties, PW_KEY_NODE_ID, "%d", node->global->id);
 	pw_properties_setf(port->properties, PW_KEY_OBJECT_ID, "%d", port->info.id);
+	pw_properties_setf(port->properties, PW_KEY_OBJECT_SERIAL, "%"PRIu64,
+			pw_global_get_serial(port->global));
 	port->info.props = &port->properties->dict;
 
 	pw_global_update_keys(port->global, &port->properties->dict, keys);
@@ -982,17 +1021,27 @@ int pw_impl_port_add(struct pw_impl_port *port, struct pw_impl_node *node)
 			pw_properties_setf(port->properties, PW_KEY_PORT_NAME, "%s_%d", dir, port->port_id);
 		}
 	}
+	if (pw_properties_get(port->properties, PW_KEY_PORT_ALIAS) == NULL) {
+		const struct pw_properties *nprops;
+		const char *node_name;
+
+		nprops = pw_impl_node_get_properties(node);
+		if ((node_name = pw_properties_get(nprops, PW_KEY_NODE_NICK)) == NULL &&
+		    (node_name = pw_properties_get(nprops, PW_KEY_NODE_DESCRIPTION)) == NULL &&
+		    (node_name = pw_properties_get(nprops, PW_KEY_NODE_NAME)) == NULL)
+			node_name = "node";
+
+		pw_properties_setf(port->properties, PW_KEY_PORT_ALIAS, "%s:%s",
+				node_name,
+				pw_properties_get(port->properties, PW_KEY_PORT_NAME));
+	}
+
 	port->info.props = &port->properties->dict;
 
 	if (control) {
 		pw_log_debug("%p: setting node control", port);
 	} else {
-		pw_log_debug("%p: setting node io", port);
-		spa_node_port_set_io(node->node,
-				     port->direction, port->port_id,
-				     SPA_IO_Buffers,
-				     &port->rt.io, sizeof(port->rt.io));
-
+		pw_log_debug("%p: setting mixer io", port);
 		spa_node_port_set_io(port->mix,
 			     pw_direction_reverse(port->direction), 0,
 			     SPA_IO_Buffers,
@@ -1067,10 +1116,12 @@ static void pw_impl_port_remove(struct pw_impl_port *port)
 	}
 
 	if (port->direction == PW_DIRECTION_INPUT) {
-		pw_map_remove(&node->input_port_map, port->port_id);
+		if ((res = pw_map_insert_at(&node->input_port_map, port->port_id, NULL)) < 0)
+			pw_log_warn("%p: can't remove input port: %s", port, spa_strerror(res));
 		node->info.n_input_ports--;
 	} else {
-		pw_map_remove(&node->output_port_map, port->port_id);
+		if ((res = pw_map_insert_at(&node->output_port_map, port->port_id, NULL)) < 0)
+			pw_log_warn("%p: can't remove output port: %s", port, spa_strerror(res));
 		node->info.n_output_ports--;
 	}
 
@@ -1148,8 +1199,8 @@ static void result_port_params(void *data, int seq, int res, uint32_t type, cons
 			d->callback(d->data, seq, r->id, r->index, r->next, r->param);
 			if (d->cache) {
 				if (d->count++ == 0)
-					pw_param_add(&impl->pending_list, r->id, NULL);
-				pw_param_add(&impl->pending_list, r->id, r->param);
+					pw_param_add(&impl->pending_list, seq, r->id, NULL);
+				pw_param_add(&impl->pending_list, seq, r->id, r->param);
 			}
 		}
 		break;
@@ -1194,7 +1245,7 @@ int pw_impl_port_for_each_param(struct pw_impl_port *port,
 	if (pi->user == 1) {
 		struct pw_param *p;
 		uint8_t buffer[1024];
-		struct spa_pod_builder b = { 0 };
+		struct spa_pod_dynamic_builder b;
 	        struct spa_result_node_params result;
 		uint32_t count = 0;
 
@@ -1202,21 +1253,23 @@ int pw_impl_port_for_each_param(struct pw_impl_port *port,
 		result.next = 0;
 
 		spa_list_for_each(p, &impl->param_list, link) {
-			result.index = result.next++;
 			if (p->id != param_id)
 				continue;
 
+			result.index = result.next++;
 			if (result.index < index)
 				continue;
 
-			spa_pod_builder_init(&b, buffer, sizeof(buffer));
-			if (spa_pod_filter(&b, &result.param, p->param, filter) != 0)
-				continue;
+			spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
 
-			pw_log_debug("%p: %d param %u", port, seq, result.index);
-			result_port_params(&user_data, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+			if (spa_pod_filter(&b.b, &result.param, p->param, filter) >= 0) {
+				pw_log_debug("%p: %d param %u", port, seq, result.index);
+				result_port_params(&user_data, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+				count++;
+			}
+			spa_pod_dynamic_builder_clean(&b);
 
-			if (++count == max)
+			if (count == max)
 				break;
 		}
 		res = 0;
@@ -1233,7 +1286,7 @@ int pw_impl_port_for_each_param(struct pw_impl_port *port,
 		spa_hook_remove(&listener);
 
 		if (user_data.cache) {
-			pw_param_update(&impl->param_list, &impl->pending_list);
+			pw_param_update(&impl->param_list, &impl->pending_list, 0, NULL);
 			pi->user = 1;
 		}
 	}
@@ -1279,7 +1332,7 @@ int pw_impl_port_for_each_filtered_param(struct pw_impl_port *in_port,
 		return res;
 
 	if (fd.n_params == 0)
-		res = do_filter(&filter, seq, 0, 0, 0, NULL);
+		res = do_filter(&fd, seq, 0, 0, 0, NULL);
 
 	return res;
 }
@@ -1311,11 +1364,14 @@ int pw_impl_port_recalc_latency(struct pw_impl_port *port)
 	struct spa_pod *param;
 	struct spa_pod_builder b = { 0 };
 	uint8_t buffer[1024];
+	bool changed;
 
 	if (port->destroying)
 		return 0;
 
-	latency = SPA_LATENCY_INFO(SPA_DIRECTION_REVERSE(port->direction));
+	/* given an output port, we calculate the total latency to the sinks or the input
+	 * latency. */
+	spa_latency_info_combine_start(&latency, SPA_DIRECTION_REVERSE(port->direction));
 
 	if (port->direction == PW_DIRECTION_OUTPUT) {
 		spa_list_for_each(l, &port->links, output_link) {
@@ -1338,20 +1394,24 @@ int pw_impl_port_recalc_latency(struct pw_impl_port *port)
 					latency.min_ns, latency.max_ns);
 		}
 	}
+	spa_latency_info_combine_finish(&latency);
 
 	current = &port->latency[latency.direction];
 
-	if (spa_latency_info_compare(current, &latency) == 0)
-		return 0;
+	changed = spa_latency_info_compare(current, &latency) != 0;
 
-	*current = latency;
-
-	pw_log_debug("port %d: set %s latency %f-%f %d-%d %"PRIu64"-%"PRIu64,
-			port->info.id, pw_direction_as_string(latency.direction),
+	pw_log_info("port %d: %s %s latency %f-%f %d-%d %"PRIu64"-%"PRIu64,
+			port->info.id, changed ? "set" : "keep",
+			pw_direction_as_string(latency.direction),
 			latency.min_quantum, latency.max_quantum,
 			latency.min_rate, latency.max_rate,
 			latency.min_ns, latency.max_ns);
 
+	if (!changed)
+		return 0;
+
+	*current = latency;
+
 	if (!port->have_latency_param)
 		return 0;
 
@@ -1419,6 +1479,10 @@ int pw_impl_port_set_param(struct pw_impl_port *port, uint32_t id, uint32_t flag
 		if (param == NULL || res < 0) {
 			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_CONFIGURE, 0, NULL);
 		}
+		else if (spa_pod_is_fixated(param) <= 0) {
+			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_CONFIGURE, 0, NULL);
+			pw_impl_port_emit_param_changed(port, id);
+		}
 		else if (!SPA_RESULT_IS_ASYNC(res)) {
 			pw_impl_port_update_state(port, PW_IMPL_PORT_STATE_READY, 0, NULL);
 		}
