@@ -125,12 +125,12 @@ struct stream {
 	uint32_t port_change_mask_all;
 	struct spa_port_info port_info;
 	struct pw_properties *port_props;
-#define IDX_EnumFormat	0
-#define IDX_Meta	1
-#define IDX_IO		2
-#define IDX_Format	3
-#define IDX_Buffers	4
-#define IDX_Latency	5
+#define PORT_EnumFormat	0
+#define PORT_Meta	1
+#define PORT_IO		2
+#define PORT_Format	3
+#define PORT_Buffers	4
+#define PORT_Latency	5
 #define N_PORT_PARAMS	6
 	struct spa_param_info port_params[N_PORT_PARAMS];
 
@@ -138,9 +138,11 @@ struct stream {
 
 	uint32_t change_mask_all;
 	struct spa_node_info info;
-#define IDX_PropInfo	0
-#define IDX_Props	1
-#define N_NODE_PARAMS	2
+#define NODE_PropInfo	0
+#define NODE_Props	1
+#define NODE_EnumFormat	2
+#define NODE_Format	3
+#define N_NODE_PARAMS	4
 	struct spa_param_info params[N_NODE_PARAMS];
 
 	uint32_t media_type;
@@ -179,9 +181,13 @@ static int get_param_index(uint32_t id)
 {
 	switch (id) {
 	case SPA_PARAM_PropInfo:
-		return IDX_PropInfo;
+		return NODE_PropInfo;
 	case SPA_PARAM_Props:
-		return IDX_Props;
+		return NODE_Props;
+	case SPA_PARAM_EnumFormat:
+		return NODE_EnumFormat;
+	case SPA_PARAM_Format:
+		return NODE_Format;
 	default:
 		return -1;
 	}
@@ -191,17 +197,17 @@ static int get_port_param_index(uint32_t id)
 {
 	switch (id) {
 	case SPA_PARAM_EnumFormat:
-		return IDX_EnumFormat;
+		return PORT_EnumFormat;
 	case SPA_PARAM_Meta:
-		return IDX_Meta;
+		return PORT_Meta;
 	case SPA_PARAM_IO:
-		return IDX_IO;
+		return PORT_IO;
 	case SPA_PARAM_Format:
-		return IDX_Format;
+		return PORT_Format;
 	case SPA_PARAM_Buffers:
-		return IDX_Buffers;
+		return PORT_Buffers;
 	case SPA_PARAM_Latency:
-		return IDX_Latency;
+		return PORT_Latency;
 	default:
 		return -1;
 	}
@@ -267,7 +273,8 @@ static struct param *add_param(struct stream *impl,
 		impl->info.change_mask |= SPA_NODE_CHANGE_MASK_PARAMS;
 		impl->params[idx].flags |= SPA_PARAM_INFO_READ;
 		impl->params[idx].user++;
-	} else if ((idx = get_port_param_index(id)) != -1) {
+	}
+	if ((idx = get_port_param_index(id)) != -1) {
 		impl->port_info.change_mask |= SPA_PORT_CHANGE_MASK_PARAMS;
 		impl->port_params[idx].flags |= SPA_PARAM_INFO_READ;
 		impl->port_params[idx].user++;
@@ -313,11 +320,12 @@ static int update_params(struct stream *impl, uint32_t id,
 }
 
 
-static inline int push_queue(struct stream *stream, struct queue *queue, struct buffer *buffer)
+static inline int queue_push(struct stream *stream, struct queue *queue, struct buffer *buffer)
 {
 	uint32_t index;
 
-	if (SPA_FLAG_IS_SET(buffer->flags, BUFFER_FLAG_QUEUED))
+	if (SPA_FLAG_IS_SET(buffer->flags, BUFFER_FLAG_QUEUED) ||
+	    buffer->id >= stream->n_buffers)
 		return -EINVAL;
 
 	SPA_FLAG_SET(buffer->flags, BUFFER_FLAG_QUEUED);
@@ -330,7 +338,13 @@ static inline int push_queue(struct stream *stream, struct queue *queue, struct
 	return 0;
 }
 
-static inline struct buffer *pop_queue(struct stream *stream, struct queue *queue)
+static inline bool queue_is_empty(struct stream *stream, struct queue *queue)
+{
+	uint32_t index;
+	return spa_ringbuffer_get_read_index(&queue->ring, &index) < 1;
+}
+
+static inline struct buffer *queue_pop(struct stream *stream, struct queue *queue)
 {
 	uint32_t index, id;
 	struct buffer *buffer;
@@ -387,6 +401,28 @@ static struct buffer *get_buffer(struct pw_stream *stream, uint32_t id)
 	return NULL;
 }
 
+static inline uint32_t update_requested(struct stream *impl)
+{
+	uint32_t index, id, res = 0;
+	struct buffer *buffer;
+	struct spa_io_rate_match *r = impl->rate_match;
+
+	if (spa_ringbuffer_get_read_index(&impl->dequeued.ring, &index) < 1)
+		return 0;
+
+	id = impl->dequeued.ids[index & MASK_BUFFERS];
+	buffer = &impl->buffers[id];
+	if (r) {
+		buffer->this.requested = r->size;
+		res = r->size > 0 ? 1 : 0;
+	} else {
+		buffer->this.requested = impl->quantum;
+		res = 1;
+	}
+	pw_log_trace_fp("%p: update buffer:%u size:%"PRIu64, impl, id, buffer->this.requested);
+	return res;
+}
+
 static int
 do_call_process(struct spa_loop *loop,
                  bool async, uint32_t seq, const void *data, size_t size, void *user_data)
@@ -398,9 +434,11 @@ do_call_process(struct spa_loop *loop,
 	return 0;
 }
 
-static void call_process(struct stream *impl)
+static inline void call_process(struct stream *impl)
 {
 	pw_log_trace_fp("%p: call process rt:%u", impl, impl->process_rt);
+	if (impl->direction == SPA_DIRECTION_OUTPUT && update_requested(impl) <= 0)
+		return;
 	if (impl->process_rt)
 		spa_callbacks_call(&impl->rt_callbacks, struct pw_stream_events, process, 0);
 	else
@@ -507,16 +545,16 @@ static int enum_params(void *object, bool is_port, int seq, uint32_t id, uint32_
 	spa_list_for_each(p, &d->param_list, link) {
 		struct spa_pod *param;
 
-		result.index = result.next++;
-		if (result.index < start)
-			continue;
-
 		param = p->param;
 		if (param == NULL || p->id != id)
 			continue;
 
 		found = true;
 
+		result.index = result.next++;
+		if (result.index < start)
+			continue;
+
 		spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
 		if (spa_pod_filter(&b.b, &result.param, param, filter) == 0) {
 			spa_node_emit_result(&d->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
@@ -550,26 +588,26 @@ static int impl_set_param(void *object, uint32_t id, uint32_t flags, const struc
 	return 0;
 }
 
-static inline uint32_t update_requested(struct stream *impl)
+static inline void copy_position(struct stream *impl, int64_t queued)
 {
-	uint32_t index, id, res = 0;
-	struct buffer *buffer;
-	struct spa_io_rate_match *r = impl->rate_match;
+	struct spa_io_position *p = impl->rt.position;
 
-	if (spa_ringbuffer_get_read_index(&impl->dequeued.ring, &index) < 1)
-		return 0;
-
-	id = impl->dequeued.ids[index & MASK_BUFFERS];
-	buffer = &impl->buffers[id];
-	if (r) {
-		buffer->this.requested = r->size;
-		res = r->size > 0 ? 1 : 0;
-	} else {
-		buffer->this.requested = impl->quantum;
-		res = 1;
+	SEQ_WRITE(impl->seq);
+	if (SPA_LIKELY(p != NULL)) {
+		impl->time.now = p->clock.nsec;
+		impl->time.rate = p->clock.rate;
+		if (SPA_UNLIKELY(impl->clock_id != p->clock.id)) {
+			impl->base_pos = p->clock.position - impl->time.ticks;
+			impl->clock_id = p->clock.id;
+		}
+		impl->time.ticks = p->clock.position - impl->base_pos;
+		impl->time.delay = 0;
+		impl->time.queued = queued;
+		impl->quantum = p->clock.duration;
 	}
-	pw_log_trace_fp("%p: update buffer:%u size:%u", impl, id, r->size);
-	return res;
+	if (SPA_LIKELY(impl->rate_match != NULL))
+		impl->rate_queued = impl->rate_match->delay;
+	SEQ_WRITE(impl->seq);
 }
 
 static int impl_send_command(void *object, const struct spa_command *command)
@@ -600,8 +638,8 @@ static int impl_send_command(void *object, const struct spa_command *command)
 			if (impl->direction == SPA_DIRECTION_INPUT)
 				impl->io->status = SPA_STATUS_NEED_DATA;
 			else if (!impl->process_rt && !impl->driving) {
-				if (update_requested(impl) > 0)
-					call_process(impl);
+				copy_position(impl, impl->queued.incount);
+				call_process(impl);
 			}
 
 			stream_set_state(stream, PW_STREAM_STATE_STREAMING, NULL);
@@ -788,7 +826,7 @@ static void clear_buffers(struct pw_stream *stream)
 	if (impl->direction == SPA_DIRECTION_INPUT) {
 		struct buffer *b;
 
-		while ((b = pop_queue(impl, &impl->dequeued))) {
+		while ((b = queue_pop(impl, &impl->dequeued))) {
 			if (b->busy)
 				ATOMIC_DEC(b->busy->count);
 		}
@@ -861,6 +899,7 @@ static int impl_port_set_param(void *object,
 	if (stream->state == PW_STREAM_STATE_ERROR)
 		return -EIO;
 
+	emit_node_info(impl, false);
 	emit_port_info(impl, false);
 
 	return 0;
@@ -883,6 +922,9 @@ static int impl_port_use_buffers(void *object,
 	if (impl->disconnecting && n_buffers > 0)
 		return -EIO;
 
+	if (n_buffers > MAX_BUFFERS)
+		return -EINVAL;
+
 	prot = PROT_READ | (direction == SPA_DIRECTION_OUTPUT ? PROT_WRITE : 0);
 
 	clear_buffers(stream);
@@ -918,6 +960,7 @@ static int impl_port_use_buffers(void *object,
 		pw_log_debug("%p: got buffer id:%d datas:%d, mapped size %d", stream, i,
 				buffers[i]->n_datas, size);
 	}
+	impl->n_buffers = n_buffers;
 
 	for (i = 0; i < n_buffers; i++) {
 		struct buffer *b = &impl->buffers[i];
@@ -927,16 +970,13 @@ static int impl_port_use_buffers(void *object,
 
 		if (impl->direction == SPA_DIRECTION_OUTPUT) {
 			pw_log_trace("%p: recycle buffer %d", stream, b->id);
-			push_queue(impl, &impl->dequeued, b);
+			queue_push(impl, &impl->dequeued, b);
 		}
 
 		SPA_FLAG_SET(b->flags, BUFFER_FLAG_ADDED);
 
 		pw_stream_emit_add_buffer(stream, &b->this);
 	}
-
-	impl->n_buffers = n_buffers;
-
 	return 0;
 }
 
@@ -945,32 +985,10 @@ static int impl_port_reuse_buffer(void *object, uint32_t port_id, uint32_t buffe
 	struct stream *d = object;
 	pw_log_trace("%p: recycle buffer %d", d, buffer_id);
 	if (buffer_id < d->n_buffers)
-		push_queue(d, &d->queued, &d->buffers[buffer_id]);
+		queue_push(d, &d->queued, &d->buffers[buffer_id]);
 	return 0;
 }
 
-static inline void copy_position(struct stream *impl, int64_t queued)
-{
-	struct spa_io_position *p = impl->rt.position;
-
-	SEQ_WRITE(impl->seq);
-	if (SPA_LIKELY(p != NULL)) {
-		impl->time.now = p->clock.nsec;
-		impl->time.rate = p->clock.rate;
-		if (SPA_UNLIKELY(impl->clock_id != p->clock.id)) {
-			impl->base_pos = p->clock.position - impl->time.ticks;
-			impl->clock_id = p->clock.id;
-		}
-		impl->time.ticks = p->clock.position - impl->base_pos;
-		impl->time.delay = 0;
-		impl->time.queued = queued;
-		impl->quantum = p->clock.duration;
-	}
-	if (SPA_LIKELY(impl->rate_match != NULL))
-		impl->rate_queued = impl->rate_match->delay;
-	SEQ_WRITE(impl->seq);
-}
-
 static int impl_node_process_input(void *object)
 {
 	struct stream *impl = object;
@@ -984,20 +1002,23 @@ static int impl_node_process_input(void *object)
 	if (io->status == SPA_STATUS_HAVE_DATA &&
 	    (b = get_buffer(stream, io->buffer_id)) != NULL) {
 		/* push new buffer */
-		if (push_queue(impl, &impl->dequeued, b) == 0) {
+		pw_log_trace_fp("%p: push %d %p", stream, b->id, io);
+		if (queue_push(impl, &impl->dequeued, b) == 0) {
 			copy_position(impl, impl->dequeued.incount);
 			if (b->busy)
 				ATOMIC_INC(b->busy->count);
 			call_process(impl);
 		}
 	}
-	if (io->status != SPA_STATUS_NEED_DATA) {
+	if (io->status != SPA_STATUS_NEED_DATA || io->buffer_id == SPA_ID_INVALID) {
 		/* pop buffer to recycle */
-		if ((b = pop_queue(impl, &impl->queued))) {
+		if ((b = queue_pop(impl, &impl->queued))) {
 			pw_log_trace_fp("%p: recycle buffer %d", stream, b->id);
-		} else if (io->status == -EPIPE)
-			return io->status;
-		io->buffer_id = b ? b->id : SPA_ID_INVALID;
+			io->buffer_id = b->id;
+		} else {
+			pw_log_trace_fp("%p: no buffers to recycle", stream);
+			io->buffer_id = SPA_ID_INVALID;
+		}
 		io->status = SPA_STATUS_NEED_DATA;
 	}
 	if (impl->driving && impl->using_trigger)
@@ -1013,28 +1034,36 @@ static int impl_node_process_output(void *object)
 	struct spa_io_buffers *io = impl->io;
 	struct buffer *b;
 	int res;
-	uint32_t index;
-	bool recycled;
+	bool ask_more;
 
 again:
 	pw_log_trace_fp("%p: process out status:%d id:%d", stream,
 			io->status, io->buffer_id);
 
-	recycled = false;
+	ask_more = false;
 	if ((res = io->status) != SPA_STATUS_HAVE_DATA) {
 		/* recycle old buffer */
 		if ((b = get_buffer(stream, io->buffer_id)) != NULL) {
 			pw_log_trace_fp("%p: recycle buffer %d", stream, b->id);
-			push_queue(impl, &impl->dequeued, b);
-			recycled = true;
+			queue_push(impl, &impl->dequeued, b);
 		}
 
 		/* pop new buffer */
-		if ((b = pop_queue(impl, &impl->queued)) != NULL) {
+		if ((b = queue_pop(impl, &impl->queued)) != NULL) {
 			impl->drained = false;
 			io->buffer_id = b->id;
 			res = io->status = SPA_STATUS_HAVE_DATA;
 			pw_log_trace_fp("%p: pop %d %p", stream, b->id, io);
+			/* we have a buffer, if we are not rt and don't follow
+			 * any rate matching and there are no more
+			 * buffers queued and there is a buffer to dequeue, ask for
+			 * more buffers so that we have one in the next round.
+			 * If we are using rate matching we need to wait until the
+			 * rate matching node (audioconvert) has been scheduled to
+			 * update the values. */
+			ask_more = !impl->process_rt && impl->rate_match == NULL &&
+				queue_is_empty(impl, &impl->queued) &&
+				!queue_is_empty(impl, &impl->dequeued);
 		} else if (impl->draining || impl->drained) {
 			impl->draining = true;
 			impl->drained = true;
@@ -1045,7 +1074,12 @@ again:
 			io->buffer_id = SPA_ID_INVALID;
 			res = io->status = SPA_STATUS_NEED_DATA;
 			pw_log_trace_fp("%p: no more buffers %p", stream, io);
+			ask_more = true;
 		}
+	} else {
+		ask_more = !impl->process_rt &&
+			queue_is_empty(impl, &impl->queued) &&
+			!queue_is_empty(impl, &impl->dequeued);
 	}
 
 	copy_position(impl, impl->queued.outcount);
@@ -1053,18 +1087,12 @@ again:
 	if (!impl->draining && !impl->driving) {
 		/* we're not draining, not a driver check if we need to get
 		 * more buffers */
-		if (!impl->process_rt && (recycled || res == SPA_STATUS_NEED_DATA)) {
-			/* not realtime and we have a free buffer, trigger process so that we have
-			 * data in the next round. */
-			if (update_requested(impl) > 0)
-				call_process(impl);
-		} else if (res == SPA_STATUS_NEED_DATA) {
-			/* realtime and we don't have a buffer, trigger process and try
-			 * again when there is something in the queue now */
-			if (update_requested(impl) > 0)
-				call_process(impl);
-			if (impl->draining ||
-			    spa_ringbuffer_get_read_index(&impl->queued.ring, &index) > 0)
+		if (ask_more) {
+			call_process(impl);
+			/* realtime, we can try again now if there is something.
+			 * non-realtime, we will have to try in the next round */
+			if (impl->process_rt &&
+			    (impl->draining || !queue_is_empty(impl, &impl->queued)))
 				goto again;
 		}
 	}
@@ -1155,7 +1183,7 @@ static int node_event_param(void *object, int seq,
 		struct control *c;
 		const struct spa_pod *type, *pod;
 		uint32_t iid, choice, n_vals, container = SPA_ID_INVALID;
-		float *vals, bool_range[3] = { 1.0, 0.0, 1.0 }, dbl[3];
+		float *vals, bool_range[3] = { 1.0f, 0.0f, 1.0f }, dbl[3];
 
 		if (spa_pod_parse_object(param,
 					SPA_TYPE_OBJECT_PropInfo, NULL,
@@ -1247,11 +1275,9 @@ static int node_event_param(void *object, int seq,
 	{
 		struct spa_pod_prop *prop;
 		struct spa_pod_object *obj = (struct spa_pod_object *) param;
-		union {
-			float f;
-			double d;
-			bool b;
-		} value;
+		float value_f;
+		double value_d;
+		bool value_b;
 		float *values;
 		uint32_t i, n_values;
 
@@ -1264,24 +1290,24 @@ static int node_event_param(void *object, int seq,
 
 			switch (c->container) {
 			case SPA_TYPE_Float:
-				if (spa_pod_get_float(&prop->value, &value.f) < 0)
+				if (spa_pod_get_float(&prop->value, &value_f) < 0)
 					continue;
 				n_values = 1;
-				values = &value.f;
+				values = &value_f;
 				break;
 			case SPA_TYPE_Double:
-				if (spa_pod_get_double(&prop->value, &value.d) < 0)
+				if (spa_pod_get_double(&prop->value, &value_d) < 0)
 					continue;
 				n_values = 1;
-				value.f = value.d;
-				values = &value.f;
+				value_f = value_d;
+				values = &value_f;
 				break;
 			case SPA_TYPE_Bool:
-				if (spa_pod_get_bool(&prop->value, &value.b) < 0)
+				if (spa_pod_get_bool(&prop->value, &value_b) < 0)
 					continue;
-				value.f = value.b ? 1.0 : 0.0;
+				value_f = value_b ? 1.0f : 0.0f;
 				n_values = 1;
-				values = &value.f;
+				values = &value_f;
 				break;
 			case SPA_TYPE_Array:
 				if ((values = spa_pod_get_array(&prop->value, &n_values)) == NULL ||
@@ -1383,7 +1409,7 @@ struct match {
 	struct pw_stream *stream;
 	int count;
 };
-#define MATCH_INIT(s) (struct match){ .stream = s }
+#define MATCH_INIT(s) ((struct match){ .stream = (s) })
 
 static int execute_match(void *data, const char *location, const char *action,
 		const char *val, size_t len)
@@ -1823,8 +1849,10 @@ pw_stream_connect(struct pw_stream *stream,
 	if (!impl->process_rt)
 		impl->info.flags |= SPA_NODE_FLAG_ASYNC;
 	impl->info.props = &stream->properties->dict;
-	impl->params[IDX_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, 0);
-	impl->params[IDX_Props] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_WRITE);
+	impl->params[NODE_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, 0);
+	impl->params[NODE_Props] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_WRITE);
+	impl->params[NODE_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, 0);
+	impl->params[NODE_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
 	impl->info.params = impl->params;
 	impl->info.n_params = N_NODE_PARAMS;
 	impl->info.change_mask = impl->change_mask_all;
@@ -1839,12 +1867,12 @@ pw_stream_connect(struct pw_stream *stream,
 	impl->port_info.flags = 0;
 	if (SPA_FLAG_IS_SET(flags, PW_STREAM_FLAG_ALLOC_BUFFERS))
 		impl->port_info.flags |= SPA_PORT_FLAG_CAN_ALLOC_BUFFERS;
-	impl->port_params[IDX_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, 0);
-	impl->port_params[IDX_Meta] = SPA_PARAM_INFO(SPA_PARAM_Meta, 0);
-	impl->port_params[IDX_IO] = SPA_PARAM_INFO(SPA_PARAM_IO, 0);
-	impl->port_params[IDX_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
-	impl->port_params[IDX_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
-	impl->port_params[IDX_Latency] = SPA_PARAM_INFO(SPA_PARAM_Latency, SPA_PARAM_INFO_WRITE);
+	impl->port_params[PORT_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, 0);
+	impl->port_params[PORT_Meta] = SPA_PARAM_INFO(SPA_PARAM_Meta, 0);
+	impl->port_params[PORT_IO] = SPA_PARAM_INFO(SPA_PARAM_IO, 0);
+	impl->port_params[PORT_Format] = SPA_PARAM_INFO(SPA_PARAM_Format, SPA_PARAM_INFO_WRITE);
+	impl->port_params[PORT_Buffers] = SPA_PARAM_INFO(SPA_PARAM_Buffers, 0);
+	impl->port_params[PORT_Latency] = SPA_PARAM_INFO(SPA_PARAM_Latency, SPA_PARAM_INFO_WRITE);
 	impl->port_info.props = &impl->port_props->dict;
 	impl->port_info.params = impl->port_params;
 	impl->port_info.n_params = N_PORT_PARAMS;
@@ -2229,7 +2257,7 @@ struct pw_buffer *pw_stream_dequeue_buffer(struct pw_stream *stream)
 	struct buffer *b;
 	int res;
 
-	if ((b = pop_queue(impl, &impl->dequeued)) == NULL) {
+	if ((b = queue_pop(impl, &impl->dequeued)) == NULL) {
 		res = -errno;
 		pw_log_trace_fp("%p: no more buffers: %m", stream);
 		errno = -res;
@@ -2240,7 +2268,7 @@ struct pw_buffer *pw_stream_dequeue_buffer(struct pw_stream *stream)
 	if (b->busy && impl->direction == SPA_DIRECTION_OUTPUT) {
 		if (ATOMIC_INC(b->busy->count) > 1) {
 			ATOMIC_DEC(b->busy->count);
-			push_queue(impl, &impl->dequeued, b);
+			queue_push(impl, &impl->dequeued, b);
 			pw_log_trace_fp("%p: buffer busy", stream);
 			errno = EBUSY;
 			return NULL;
@@ -2260,7 +2288,7 @@ int pw_stream_queue_buffer(struct pw_stream *stream, struct pw_buffer *buffer)
 		ATOMIC_DEC(b->busy->count);
 
 	pw_log_trace_fp("%p: queue buffer %d", stream, b->id);
-	if ((res = push_queue(impl, &impl->queued, b)) < 0)
+	if ((res = queue_push(impl, &impl->queued, b)) < 0)
 		return res;
 
 	if (impl->direction == SPA_DIRECTION_OUTPUT &&
@@ -2281,9 +2309,9 @@ do_flush(struct spa_loop *loop,
 
 	pw_log_trace_fp("%p: flush", impl);
 	do {
-		b = pop_queue(impl, &impl->queued);
+		b = queue_pop(impl, &impl->queued);
 		if (b != NULL)
-			push_queue(impl, &impl->dequeued, b);
+			queue_push(impl, &impl->dequeued, b);
 	}
 	while (b);
 
@@ -2309,7 +2337,7 @@ int pw_stream_flush(struct pw_stream *stream, bool drain)
 	struct stream *impl = SPA_CONTAINER_OF(stream, struct stream, this);
 	pw_loop_invoke(impl->context->data_loop,
 			drain ? do_drain : do_flush, 1, NULL, 0, true, impl);
-	if (!drain)
+	if (!drain && impl->node != NULL)
 		spa_node_send_command(impl->node->node,
 				&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Flush));
 	return 0;
@@ -2330,7 +2358,7 @@ do_trigger_process(struct spa_loop *loop,
 	int res;
 	if (impl->direction == SPA_DIRECTION_OUTPUT) {
 		if (impl->process_rt)
-			spa_callbacks_call(&impl->rt_callbacks, struct pw_stream_events, process, 0);
+			call_process(impl);
 		res = impl->node_methods.process(impl);
 	} else {
 		res = SPA_STATUS_NEED_DATA;
@@ -2364,11 +2392,9 @@ int pw_stream_trigger_process(struct pw_stream *stream)
 	if (!impl->driving && !impl->trigger) {
 		res = trigger_request_process(impl);
 	} else {
-		if (impl->direction == SPA_DIRECTION_OUTPUT &&
-		    !impl->process_rt) {
-			pw_loop_invoke(impl->context->main_loop,
-				do_call_process, 1, NULL, 0, false, impl);
-		}
+		if (!impl->process_rt)
+			call_process(impl);
+
 		res = pw_loop_invoke(impl->context->data_loop,
 			do_trigger_process, 1, NULL, 0, false, impl);
 	}
